/*=============================================================================
	Module Name:
			ConRedir.cpp

	Abstract:
			Console application standard handle redirector

    License:
            GNU General Public License (GPL)

	Revision History:
			* Nov 20, 2004:  Derived from QuickWin by Attila Afra
            * Dec 16, 2004:  Child process termination bug fixed
            * Feb 27, 2005:  Custom priority class
            * Mar 25, 2005:  Line based output events

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

// Warnings
#pragma warn -8004  // ... is assigned a value that is never used
#pragma warn -8060  // Possible incorrect assignment

// Includes
#include "ConRedir.h"

// Constants
#define BUFFER_SIZE     1024     // size of the output buffer

// API function
BOOL WINAPI IsWinNT()
{
    // Get Windows version
    DWORD WindowsVersion      = GetVersion();
    DWORD WindowsMajorVersion = (DWORD)(LOBYTE(LOWORD(WindowsVersion)));
    DWORD WindowsMinorVersion = (DWORD)(HIBYTE(LOWORD(WindowsVersion)));

    // Running on Win9x ?
    if (WindowsVersion >= 0x80000000) return FALSE;

    // Running on WinNT
    return TRUE;
}

/*=============================================================================
    CConRedir class
=============================================================================*/

CConRedir::CConRedir()
{
	// Initialization
	m_hStdIn         = NULL;
	m_hStdOut        = NULL;
	m_hStdErr        = NULL;
	m_hStdInWrite    = NULL;
	m_hStdOutRead    = NULL;
	m_hStdErrRead    = NULL;
	m_hChildProcess  = NULL;
	m_hStdOutThread  = NULL;
	m_hStdErrThread  = NULL;
	m_hProcessThread = NULL;
	m_hExitEvent     = NULL;
	m_bRunThread     = FALSE;

    m_StdOutLineBufSize = BUFFER_SIZE * 2;
    m_StdOutLineBuf = (CHAR *)malloc(m_StdOutLineBufSize * sizeof(CHAR));
    m_StdErrLineBufSize = BUFFER_SIZE * 2;
    m_StdErrLineBuf = (CHAR *)malloc(m_StdErrLineBufSize * sizeof(CHAR));
}

CConRedir::~CConRedir()
{
	TerminateChildProcess();
}

/*-----------------------------------------------------------------------------
    CConRedir >> Operations
-----------------------------------------------------------------------------*/

// Create standard handles, try to start child from command line
BOOL CConRedir::StartChildProcess(LPCSTR lpszCmdLine, DWORD dwPriorityClass, BOOL bShowChildWindow)
{
	HANDLE hProcess = GetCurrentProcess();

	// Set up the security attributes struct
	SECURITY_ATTRIBUTES sa;

	ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;

	HANDLE hStdInWriteTmp, hStdOutReadTmp, hStdErrReadTmp;

	// Create the child stdin pipe
	if (!CreatePipe(&m_hStdIn, &hStdInWriteTmp, &sa, 0))
        return FALSE;

	// Create the child stdout pipe
	if (!CreatePipe(&hStdOutReadTmp, &m_hStdOut, &sa, 0))
        return FALSE;

	// Create the child stderr pipe
	if (!CreatePipe(&hStdErrReadTmp, &m_hStdErr, &sa, 0))
        return FALSE;

	// Create new stdin write, stdout and stderr read handles.
	// Set the properties to FALSE. Otherwise, the child inherits the
	// properties and, as a result, non-closeable handles to the pipes
	// are created.
	if (!DuplicateHandle(hProcess, hStdInWriteTmp,
		hProcess, &m_hStdInWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
        return FALSE;

	if (!DuplicateHandle(hProcess, hStdOutReadTmp,
		hProcess, &m_hStdOutRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
        return FALSE;

	if (!DuplicateHandle(hProcess, hStdErrReadTmp,
		hProcess, &m_hStdErrRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
        return FALSE;

	// Close inheritable copies of the handles you do not want to be inherited
	CloseHandle(hStdInWriteTmp);
	CloseHandle(hStdOutReadTmp);
	CloseHandle(hStdErrReadTmp);

	// Start child process with redirected stdout, stdin & stderr
	m_hChildProcess = PrepAndLaunchRedirectedChild(lpszCmdLine,
		m_hStdOut, m_hStdIn, m_hStdErr, dwPriorityClass, bShowChildWindow);

	if (m_hChildProcess == NULL)
    {
        // Close all handles and return FALSE
		CloseHandle(m_hStdIn);
		m_hStdIn = NULL;

		CloseHandle(m_hStdOut);
		m_hStdOut = NULL;

		CloseHandle(m_hStdErr);
		m_hStdErr = NULL;

		return FALSE;
    }

	DWORD dwThreadID;
	m_bRunThread = TRUE;

	// Create Exit event
	m_hExitEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (m_hExitEvent == NULL)
        return FALSE;

	// Launch the thread that read the child stdout
    m_StdOutLineBuf[0] = '\0';
    
	m_hStdOutThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)staticStdOutThread,
		(LPVOID)this, 0, &dwThreadID);

	if (m_hStdOutThread == NULL)
        return FALSE;

	// Launch the thread that reads the child stderr
    m_StdErrLineBuf[0] = '\0';

	m_hStdErrThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)staticStdErrThread,
		(LPVOID)this, 0, &dwThreadID);

	if (m_hStdErrThread == NULL)
        return FALSE;

	// Launch the thread that monitors the child process
	m_hProcessThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)staticProcessThread,
		(LPVOID)this, 0, &dwThreadID);

	if (m_hProcessThread == NULL)
        return FALSE;

	// Virtual function to notify derived class that the child is started
	OnChildStarted(lpszCmdLine);

	return TRUE;
}

// Check if the child process is running
// On NT/2000 the handle must have PROCESS_QUERY_INFORMATION access
BOOL CConRedir::IsChildRunning() const
{
	DWORD dwExitCode;

    if (m_hChildProcess == NULL)
        return FALSE;

	GetExitCodeProcess(m_hChildProcess, &dwExitCode);
	return (dwExitCode == STILL_ACTIVE) ? TRUE : FALSE;
}

// Changes priority class of the child process
void CConRedir::SetPriorityClass(DWORD dwPriorityClass)
{
    if (IsChildRunning())
        ::SetPriorityClass(m_hChildProcess, dwPriorityClass);
}

// Terminate child process
void CConRedir::TerminateChildProcess()
{
	// Tell the threads to exit and wait for process thread to die
	m_bRunThread = FALSE;
	SetEvent(m_hExitEvent);
	//Sleep(500);

    // Stop the child process if not already stopped
	if (IsChildRunning())
    {
		TerminateProcess(m_hChildProcess, 1);
		WaitForSingleObject(m_hChildProcess, 1000);
    }
	m_hChildProcess = NULL;

	// Check the process thread
	if (m_hProcessThread != NULL)
	{
		WaitForSingleObject(m_hProcessThread, 1000);
		m_hProcessThread = NULL;
	}

	// Close all child handles first
	if (m_hStdIn != NULL)
		CloseHandle(m_hStdIn);
	m_hStdIn = NULL;

	if (m_hStdOut != NULL)
		CloseHandle(m_hStdOut);
	m_hStdOut = NULL;

	if (m_hStdErr != NULL)
		CloseHandle(m_hStdErr);
	m_hStdErr = NULL;

    //Sleep(100);

	// Close all parent handles.
	if (m_hStdInWrite != NULL)
		CloseHandle(m_hStdInWrite);
	m_hStdInWrite = NULL;

	if (m_hStdOutRead != NULL)
		CloseHandle(m_hStdOutRead);
	m_hStdOutRead = NULL;

	if (m_hStdErrRead != NULL)
		CloseHandle(m_hStdErrRead);
	m_hStdErrRead = NULL;

    //Sleep(100);

    // Stop the stdout read thread
	if (m_hStdOutThread != NULL)
	{
		if (!IsWinNT())
			TerminateThread(m_hStdOutThread, 1);
		WaitForSingleObject(m_hStdOutThread, 1000);
		m_hStdOutThread = NULL;
	}

    // Stop the stderr read thread
	if (m_hStdErrThread != NULL)
	{
		if (!IsWinNT())
			TerminateThread(m_hStdErrThread, 1);
		WaitForSingleObject(m_hStdErrThread, 1000);
		m_hStdErrThread = NULL;
	}

    //Sleep(100);

	// Cleanup the exit event
	if (m_hExitEvent != NULL)
		CloseHandle(m_hExitEvent);
	m_hExitEvent = NULL;
}

// Function that writes to the child stdin
BOOL CConRedir::WriteChildStdIn(LPCSTR lpszInput)
{
	DWORD nBytesWrote;
	DWORD Length = strlen(lpszInput);

	if (m_hStdInWrite != NULL && Length > 0)
	{
		if (!WriteFile(m_hStdInWrite, lpszInput, Length, &nBytesWrote, NULL))
		{
			if (GetLastError() == ERROR_NO_DATA)
				;				// pipe was closed (do nothing)
			else
				return FALSE;	// something bad happened
		}
	}

    return TRUE;
}

/*-----------------------------------------------------------------------------
    CConRedir >> Internal functions
-----------------------------------------------------------------------------*/

// Launch the process that you want to redirect
HANDLE CConRedir::PrepAndLaunchRedirectedChild(LPCSTR lpszCmdLine,
	HANDLE hStdOut, HANDLE hStdIn, HANDLE hStdErr,
	DWORD dwPriorityClass, BOOL bShowChildWindow)
{
	HANDLE hProcess = GetCurrentProcess();

	PROCESS_INFORMATION pi;

	// Set up the startup info struct
	STARTUPINFO si;
	ZeroMemory(&si, sizeof(STARTUPINFO));
	si.cb = sizeof(STARTUPINFO);
	si.dwFlags    = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
	si.hStdOutput = hStdOut;
	si.hStdInput  = hStdIn;
	si.hStdError  = hStdErr;

	// Use this if you want to show the child.
    // Note that dwFlags must include STARTF_USESHOWWINDOW if you want to
	// use the wShowWindow flags.
	si.wShowWindow = bShowChildWindow ? SW_SHOW: SW_HIDE;

	// Create the NULL security token for the process
	LPVOID lpSD = NULL;
	LPSECURITY_ATTRIBUTES lpSA = NULL;

	// On NT/2000 the handle must have PROCESS_QUERY_INFORMATION access.
	// This is made using an empty security descriptor. It is not the same
	// as using a NULL pointer for the security attribute!
	if (IsWinNT())
	{
		lpSD = GlobalAlloc(GPTR, SECURITY_DESCRIPTOR_MIN_LENGTH);
		InitializeSecurityDescriptor(lpSD, SECURITY_DESCRIPTOR_REVISION);
		SetSecurityDescriptorDacl(lpSD, -1, 0, 0);

		lpSA = (LPSECURITY_ATTRIBUTES)GlobalAlloc(GPTR, sizeof(SECURITY_ATTRIBUTES));
		lpSA->nLength = sizeof(SECURITY_ATTRIBUTES);
		lpSA->lpSecurityDescriptor = lpSD;
		lpSA->bInheritHandle = TRUE;
	}

	// Try to spawn the process
	BOOL bResult = CreateProcess(NULL, (char *)lpszCmdLine, lpSA, NULL, TRUE,
                       CREATE_NEW_CONSOLE | dwPriorityClass, NULL, NULL, &si, &pi);

	// Cleanup memory allocation
	if (lpSA != NULL)
		GlobalFree(lpSA);

	if (lpSD != NULL)
		GlobalFree(lpSD);

	// Return if an error occurs
	if (!bResult)
        return FALSE;

	// Close any unnecessary handles
    CloseHandle(pi.hThread);

	// Save global child process handle to cause threads to exit
	return pi.hProcess;
}

// Variable initialization
BOOL CConRedir::m_bRunThread = TRUE;

// Thread to read the child stdout
int CConRedir::StdOutThread(HANDLE hStdOutRead)
{
	DWORD  nBytesRead;
	CHAR   lpszBuffer[BUFFER_SIZE + 1];
    CHAR  *p, *pn;

	while (m_bRunThread)
	{
		if (!ReadFile(hStdOutRead, lpszBuffer, BUFFER_SIZE,
			&nBytesRead, NULL) || !nBytesRead)
		{
			if (GetLastError() == ERROR_BROKEN_PIPE)
				break;			// pipe done - normal exit path.
			else
				return -1;   	// something bad happened
		}

		if (nBytesRead)
		{
			// Virtual function to notify derived class that
			// characters are writted to stdout
			lpszBuffer[nBytesRead] = '\0';
			//OnChildStdOutWrite(lpszBuffer);

            // Append the new characters to the line buffer
            if (strlen(m_StdOutLineBuf) + strlen(lpszBuffer) >= m_StdOutLineBufSize)
            {
                // Buffer overflow
                m_StdOutLineBufSize += BUFFER_SIZE * 2;
                m_StdOutLineBuf = (CHAR *)realloc(m_StdOutLineBuf, m_StdOutLineBufSize * sizeof(CHAR));
            }

            strcat(m_StdOutLineBuf, lpszBuffer);

            // Virtual function to notify derived class that
            // new lines are writted to stdout
            p = m_StdOutLineBuf;
            while (*p && (pn = strchr(p, '\r')))
            {
                *pn = '\0';
                if (*p == '\n')
                    p++;

                OnChildStdOutWriteLn(p);
                p = pn + 1;
            }

            // Remove entire lines from the buffer
            if (*p == '\n')
                p++;
            strcpy(m_StdOutLineBuf, p);
		}
	}

	return 0;
}

// Thread to read the child stderr
int CConRedir::StdErrThread(HANDLE hStdErrRead)
{
	DWORD  nBytesRead;
	CHAR   lpszBuffer[BUFFER_SIZE + 1];
    CHAR  *p, *pn;

	while (m_bRunThread)
	{
		if (!ReadFile(hStdErrRead, lpszBuffer, BUFFER_SIZE,
			&nBytesRead, NULL) || !nBytesRead)
		{
			if (GetLastError() == ERROR_BROKEN_PIPE)
				break;			// pipe done - normal exit path
			else
				return -1;  	// something bad happened.
		}

		if (nBytesRead)
		{
			// Virtual function to notify derived class that
			// characters are writted to stderr
			lpszBuffer[nBytesRead] = '\0';
			//OnChildStdErrWrite(lpszBuffer);

            // Append the new characters to the line buffer
            if (strlen(m_StdErrLineBuf) + strlen(lpszBuffer) >= m_StdErrLineBufSize)
            {
                // Buffer overflow
                m_StdErrLineBufSize += BUFFER_SIZE * 2;
                m_StdErrLineBuf = (CHAR *)realloc(m_StdErrLineBuf, m_StdErrLineBufSize * sizeof(CHAR));
            }

            strcat(m_StdErrLineBuf, lpszBuffer);

            // Virtual function to notify derived class that
            // new lines are writted to stdout
            p = m_StdErrLineBuf;
            while (*p && (pn = strchr(p, '\r')))
            {
                *pn = '\0';
                if (*p == '\n')
                    p++;

                OnChildStdErrWriteLn(p);
                p = pn + 1;
            }

            // Remove entire lines from the buffer
            if (*p == '\n')
                p++;
            strcpy(m_StdErrLineBuf, p);
		}
	}

	return 0;
}

// Thread to monitor the child process
int CConRedir::ProcessThread()
{
	HANDLE hWaitHandle[2];
	hWaitHandle[0] = m_hExitEvent;
	hWaitHandle[1] = m_hChildProcess;

	while (m_bRunThread)
	{
		switch (WaitForMultipleObjects(2, hWaitHandle, FALSE, 1))
		{
			case WAIT_OBJECT_0 + 0:	// exit on event
                if (m_bRunThread == TRUE)
                    return -1;
				break;

			case WAIT_OBJECT_0 + 1:	// child process exit
				if (m_bRunThread == FALSE)
                    return -1;
				m_bRunThread = FALSE;
				break;
		}
	}

    // Flush line buffers
    WaitForSingleObject(m_hStdOutThread, 100);
    if (*m_StdOutLineBuf)
    {
        OnChildStdOutWriteLn(m_StdOutLineBuf);
        *m_StdOutLineBuf = '\0';
    }
    //free(m_StdOutLineBuf);
    //m_StdOutLineBuf = NULL;

    WaitForSingleObject(m_hStdErrThread, 100);
    if (*m_StdErrLineBuf)
    {
        OnChildStdErrWriteLn(m_StdErrLineBuf);
        *m_StdErrLineBuf = '\0';
    }
    //free(m_StdErrLineBuf);
    //m_StdErrLineBuf = NULL;

	// Virtual function to notify derived class that child process is terminated.
	// Application must call TerminateChildProcess() but not direcly from this thread!
	OnChildTerminate();
	return 0;
}
