/*=============================================================================
    Project Name:
            FilmShrink

	Module Name:
			FSKGlobal.cpp

	Abstract:
			FilmShrink global declarations

    License:
            GNU General Public License (GPL)

	Revision History:
			* Aug 10, 2004:  Created by Attila Afra
            * Jan 12, 2004:  Single instance restriction
            * Mar  4, 2005:  FileSize bug fixed
            * Mar 24, 2005:  Advanced directory tree removing function

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

// Includes
#include "avs.h"

#include "FSKSplash.h"
#include "FSKMenu.h"
#include "FSKNew.h"
#include "FSKGlobal.h"

/*-----------------------------------------------------------------------------
    Main
-----------------------------------------------------------------------------*/

HANDLE FSKAuxEvent;

// FilmShrink startup code
bool FSKStartup()
{
    const char szEventName[] = "FilmShrinkAuxEvent";

    // Check for the event
    if (OpenEvent(EVENT_ALL_ACCESS, true, szEventName))
        return false;

    // Create the event if it doesn't exist
    FSKAuxEvent = CreateEvent(NULL, false, false, szEventName);
    return true;
}

// Initialize FilmShrink
bool FSKInitialize()
{
    // Initialize path and title
    FSKInitPath();

    // Try to init registry key
    if (!FSKInitReg())
    {
        // Error
        Application->MessageBoxA("Unable to initialize registry entries.", FSK_PRODUCT_NAME, MB_ICONERROR | MB_OK);
        return false;
    }

    // Try to install AviSynth
    if (!AVSInstall(ExtractFilePath(Application->ExeName).c_str()))
    {
        // Error
        Application->MessageBoxA("Unable to initialize frameserver. Try to reinstall FilmShrink.", FSK_PRODUCT_NAME, MB_ICONERROR | MB_OK);
        return false;
    }

    // Success
    return true;
}

// Terminate FilmShrink
void FSKTerminate()
{
    // Uninstall AVS
    AVSUninstall();

    // Close registry key
    FSKCloseReg();

    // Close auxiliary handle
    CloseHandle(FSKAuxEvent);
}

/*-----------------------------------------------------------------------------
    Registry
-----------------------------------------------------------------------------*/

// VCL registry class definition
TRegistry *g_Reg;

// Open/initialize registry keys
bool FSKInitReg()
{
    g_Reg = new TRegistry();
    g_Reg->RootKey = HKEY_LOCAL_MACHINE;

    return FSKVerifyReg();
}

// Verify/fix registry entries
bool FSKVerifyReg()
{
    TStringList *Jobs;         // job index strings
    int          JobCount;     // number of valid jobs

    // Open/create FilmShrink's main registry key
    if (!g_Reg->OpenKey(FSK_REGKEY, true))
    {
        g_Reg->Free();
        return false;
    }

    // Close key
    g_Reg->CloseKey();

    // Open the jobs key
    if (!g_Reg->OpenKey(FSK_JOBS_REGKEY, true))
    {
        g_Reg->Free();
        return false;
    }

    // Get jobs
    Jobs = new TStringList();
    g_Reg->GetKeyNames(Jobs);
    g_Reg->CloseKey();

    // Process job entries
    if (Jobs->Count)
    {
        AnsiString   JobKeyName; // current jobkey
        bool         IsValid;    // is the current key a valid job?
        AnsiString   NewName;    // new key name
        TStringList *SubKeys;    // subkeys (that must be deleted)

        // Sort entries
        Jobs->Sort();

        // Check entries
        JobCount = 0;

        for (int i = 0; i < Jobs->Count; i++)
        {
            JobKeyName = AnsiString(FSK_JOBS_REGKEY) + "\\" + Jobs->Strings[i];

            // Is it valid?
            if (!g_Reg->OpenKey(JobKeyName, false))
                // Registry manipulation detected
                continue;

            // Remove subkeys
            SubKeys = new TStringList();
            g_Reg->GetKeyNames(SubKeys);

            for (int j = 0; j < SubKeys->Count; j++)
                if (g_Reg->KeyExists(SubKeys->Strings[j]))
                    g_Reg->DeleteKey(SubKeys->Strings[j]);

            // Check for necessary values
            IsValid = g_Reg->ValueExists(FSK_JOB_PATH_REGVAL) && g_Reg->GetDataType(FSK_JOB_PATH_REGVAL) == rdString &&
                      g_Reg->ValueExists(FSK_JOB_NAME_REGVAL) && g_Reg->GetDataType(FSK_JOB_NAME_REGVAL) == rdString &&
                      g_Reg->ValueExists(FSK_JOB_DATE_REGVAL) && g_Reg->GetDataType(FSK_JOB_DATE_REGVAL) == rdString &&
                      g_Reg->ValueExists(FSK_JOB_DONE_REGVAL) && g_Reg->GetDataType(FSK_JOB_DONE_REGVAL) == rdInteger;

            g_Reg->CloseKey();
            delete SubKeys;

            if (!IsValid)
            {
                if (g_Reg->KeyExists(JobKeyName))
                    g_Reg->DeleteKey(JobKeyName);
                continue;
            }

            // Rename key if necessary
            if (!g_Reg->OpenKey(FSK_JOBS_REGKEY, false))
            {
                g_Reg->Free();
                delete Jobs;
                return false;
            }

            NewName.printf("%d", JobCount);

            if (Jobs->Strings[i] != NewName)
            {
                if (!g_Reg->KeyExists(Jobs->Strings[i]))
                    // Registry manipulation detected
                    continue;

                g_Reg->MoveKey(Jobs->Strings[i], NewName, true);
            }

            g_Reg->CloseKey();
            JobCount++;
        }
    }
    else
        JobCount = 0;

    // Write number of jobs
    delete Jobs;

    if (!g_Reg->OpenKey(FSK_JOBS_REGKEY, false))
    {
        g_Reg->Free();
        return false;
    }

    if (g_Reg->ValueExists(FSK_JOBS_COUNT_REGVAL) && g_Reg->GetDataType(FSK_JOBS_COUNT_REGVAL) != rdString)
        g_Reg->DeleteValue(FSK_JOBS_COUNT_REGVAL);
    g_Reg->WriteInteger(FSK_JOBS_COUNT_REGVAL, JobCount);

    g_Reg->CloseKey();

    // Success
    return true;
}

// Close registry keys
void FSKCloseReg()
{
    g_Reg->Free();
}

/*-----------------------------------------------------------------------------
    Paths
-----------------------------------------------------------------------------*/

// String class definitions
AnsiString FSKAppTitle;
AnsiString FSKAppPath;
AnsiString FSKEncoderPath;    

// Initialize application path and title
void FSKInitPath()
{
    FSKAppTitle    = Application->Title;
    FSKAppPath     = ExtractFilePath(Application->ExeName) + "\\";
    FSKEncoderPath = FSKAppPath + FSK_PRODUCER_PATH;
}

/*-----------------------------------------------------------------------------
    Forms
-----------------------------------------------------------------------------*/

// Show the splash form
void FSKShowSplashForm()
{
    // Create new form object (if required)
    FSKSplashForm = new TFSKSplashForm(Application);

    // Show it
    FSKSplashForm->Show();

    // FilmShrink startup
    FSKSplashForm->FSKStartup();
}

// Show the main menu
void FSKShowMenuForm()
{
    // Create new form object
    FSKMenuForm = new TFSKMenuForm(Application);

    // Show it
    FSKMenuForm->Show();
}

// Show the "New Job" wizard
void FSKShowNewForm()
{
    // Create new form object
    FSKNewForm = new TFSKNewForm(FSKMenuForm);

    // Show it
    FSKNewForm->Show();
}

/*=============================================================================
    Utilities
=============================================================================*/

// Remove an entire directory tree
bool RemoveDirTree(AnsiString Dir)
{
    TSearchRec sr;
    AnsiString DirPath;
    int Result;
    bool Return = true;

    // Store the directory's path
    if (Dir[Dir.Length()] == '\\')
        DirPath = Dir;
    else
        DirPath = Dir + "\\";

    // Remove all files from directory
    Result = FindFirst(DirPath + "*.*", faAnyFile ^ faDirectory ^ faVolumeID, sr);

    while (!Result)
    {
        // Remove read-only attribute, if present
        if (sr.Attr & faReadOnly)
        {
            if (!FileSetReadOnly(DirPath + sr.Name, false))
            {
                Return = false;
                goto NextFile;
            }
        }

        // Delete current file
        if (!DeleteFile(DirPath + sr.Name))
        {
            Return = false;
            goto NextFile;
        }

NextFile:
        // Jump to next file
        Result = FindNext(sr);
    }

    FindClose(sr);

    // Remove all child directory trees
    Result = FindFirst(DirPath + "*.*", faAnyFile ^ faVolumeID, sr);

    while (!Result)
    {
        // We search only for directories
        if ((sr.Attr & faDirectory) && (sr.Name != ".") && (sr.Name != ".."))
        {
            // Remove read-only attribute, if present
            if (sr.Attr & faReadOnly)
            {
                if (!FileSetReadOnly(DirPath + sr.Name, false))
                {
                    Return = false;
                    goto NextDir;
                }
            }

            // Remove current directory tree
            if (!RemoveDirTree(DirPath + sr.Name))
            {
                Return = false;
                goto NextDir;
            }
        }

NextDir:
        // Jump to next directory
        Result = FindNext(sr);
    }

    FindClose(sr);

    // Remove the empty initial directory
    if (Return)
        RemoveDir(Dir);

    return Return;
}

// Get the size of a file
__int64 FileSize(AnsiString FileName)
{
    int Handle;
    __int64 Size;

    if ((Handle = FileOpen(FileName, fmOpenRead | fmShareDenyNone)) == -1)
        return -1;

    Size = FileSeek(Handle, __int64(0), 2);
    FileClose(Handle);

    return Size;
}