/*=============================================================================
    Project Name:
            FilmShrink

	Module Name:
			FSKMenu.cpp

	Abstract:
			FilmShrink main menu form

    License:
            GNU General Public License (GPL)

	Revision History:
			* Aug  9, 2004:  Created by Attila Afra
            * Jan 14, 2005:  About box
            * Jan 17, 2005:  Minor changes
            * Jan 23, 2005:  Help extensions
            * Jan 27, 2005:  Bugfixes
            * Feb  3, 2005:  Shutdown bug fixed
            * Feb  6, 2005:  Better time estimation
            * Mar  4, 2005:  Progress percent precision increased
                             Code cleanup
            * Mar 25, 2005:  Encoder message parser fix

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

#include <vcl.h>
#pragma hdrstop

// Warnings
#pragma warn -8060  // Possibly incorrect assignment
#pragma warn -8004  // ... is assigned a value that is never used


// Includes
#include <stdio.h>
#include <math.h>
#include "DateUtils.hpp"
#include "Shell.h"
#include "Shutdown.h"
#include "FSKGlobal.h"
#include "FSKTemplate.h"
#include "FSKMenu.h"
#include "FSKSplash.h"

// VCL stuff
#pragma package(smart_init)
#pragma link "TB2Dock"
#pragma link "TB2Toolbar"
#pragma link "TBX"
#pragma link "TB2Item"
#pragma link "ThemeMgr"
#pragma resource "*.dfm"

/*=============================================================================
    TFSKMenuForm
=============================================================================*/

// MenuForm definition
TFSKMenuForm *FSKMenuForm = NULL;

/*-----------------------------------------------------------------------------
    TFSKMenuForm >> Constructors
-----------------------------------------------------------------------------*/

// Default constructor
__fastcall TFSKMenuForm::TFSKMenuForm(TComponent* Owner) : TForm(Owner)
{
    FormCaption = Caption;

    // Initialization
    JobListSelectItemUpdateOnly = false;

    // Create encoder class
    Encoder = new TFSKEncoder();

    // Set shutdown options
    ShutCheck->Checked = false;
    ShutCheckClick(NULL);
}

// Destructor
__fastcall TFSKMenuForm::~TFSKMenuForm()
{
    // Delete encoder class
    delete Encoder;
}

/*-----------------------------------------------------------------------------
    TFSKMenuForm >> Events
-----------------------------------------------------------------------------*/

// TMenuForm::WndProc
MAIN_WNDPROC(TFSKMenuForm, FSKMainForm)


void __fastcall TFSKMenuForm::FormActivate(TObject *Sender)
{
    FSKActiveForm = this;
}

void __fastcall TFSKMenuForm::FormShow(TObject *Sender)
{
    MAIN_SHOW_INIT(FSKMainForm);

    IsWorking = false;

    // Disable job operations
    StartItem->Enabled = false;
    DelItem->Enabled   = false;
    UpItem->Enabled    = false;
    DownItem->Enabled  = false;

    // Disable abort
    AbortItem->Enabled = false;

    // Set default encoder priority
    PriorityCombo->ItemIndex = 1;
    PrevPriority = PriorityCombo->ItemIndex;

    // Reset progress info
    ResetProgInfo();

    // Load job list
    if (!FSKVerifyReg())
    {
        // Error
        RegError();
        return;
    }

    LoadJobList();

    // Set batch start button
    StartAllItem->Enabled = PendingJobCount > 0;
}

void __fastcall TFSKMenuForm::FormHide(TObject *Sender)
{
    // Save job list
    SaveJobList();
}

void __fastcall TFSKMenuForm::FormClose(TObject *Sender, TCloseAction &Action)
{
    if (IsWorking)
    {
        if (Application->MessageBoxA("Do you really want to abort the transcoding process?", FSK_PRODUCT_NAME, MB_YESNO | MB_ICONEXCLAMATION) == mrYes)
        {
            AbortJob();
            //Sleep(100);
            FSKTermFlag = true;
            Action = caFree;
            Application->Terminate();
        }
        else
        {
            FSKTermFlag = false;
            Action = caNone;
        }
    }
    else
    {
        //Sleep(100);
        FSKTermFlag = true;
        Action = caFree;
        Application->Terminate();
    }
}

/*---------------------------------------------------------------------------*/

void __fastcall TFSKMenuForm::JobListSelectItem(TObject *Sender, TListItem *Item, bool Selected)
{
    int        VolDone, VolTotal;  // number of volumes done / total number of volumes
    double     TotalProg;          // total progress
    AnsiString Caption;            // caption

    // Exit if a job is being encoded
    if (IsWorking)
    {
        JobListSelectItemUpdateOnly = false;
        return;
    }

    if (Selected)
    {
        if (!JobListSelectItemUpdateOnly)
        {
            // Enable job operations
            if (Item->StateIndex == JS_Pending)
                StartItem->Enabled = true;
            else
                StartItem->Enabled = false;

            DelItem->Enabled = true;

            if (Item->Index > 0)
                UpItem->Enabled = true;

            if (Item->Index < JobList->Items->Count - 1)
                DownItem->Enabled  = true;
        }

        // Set progress info
        ProgJobLabel->Caption   = JobList->Items->Item[Item->Index]->SubItems->Strings[JI_Name];
        ProgDateLabel->Caption  = JobList->Items->Item[Item->Index]->SubItems->Strings[JI_Date];
        ProgTimeLabel->Caption  = JobList->Items->Item[Item->Index]->SubItems->Strings[JI_Time];
        ProgLeftLabel->Caption  = JobList->Items->Item[Item->Index]->SubItems->Strings[JI_TimeLeft];


        if (JobList->Items->Item[Item->Index]->SubItems->Strings[JI_Vol] != "--")
        {
            sscanf(JobList->Items->Item[Item->Index]->SubItems->Strings[JI_Vol].c_str(), "%d / %d", &VolDone, &VolTotal);

            if (VolDone != VolTotal)
            {
                Caption.printf("Volume %d / %d:", VolDone + 1, VolTotal);
                ProgVolLabel2->Caption  = "0.0 %";
                ProgVolBar->Position    = 0;
            }
            else
            {
                Caption.printf("Volume %d / %d:", VolDone, VolTotal);
                ProgVolLabel2->Caption  = "100.0 %";
                ProgVolBar->Position    = 100;
            }
        }
        else
        {
            if (Item->StateIndex == JS_Done)
            {
                VolDone = 1;

                ProgVolLabel2->Caption  = "100.0 %";
                ProgVolBar->Position    = 100;
            }
            else
            {
                VolDone = 0;

                ProgVolLabel2->Caption  = "0.0 %";
                ProgVolBar->Position    = 0;
            }

            VolTotal = 1;
            Caption.printf("Volume:");
        }

        ProgVolLabel1->Caption  = Caption;
        TotalProg = double(VolDone * 100) / VolTotal;
        Caption.printf("%.1f %%", TotalProg);
        ProgTotalLabel2->Caption = Caption;
        ProgTotalBar->Position   = int(TotalProg);

        if (!JobListSelectItemUpdateOnly)
        {
            ProgLabel1->Enabled      = false;
            ProgLabel2->Enabled      = false;
            ProgLabel3->Enabled      = false;
            ProgLabel4->Enabled      = false;
            ProgTotalLabel1->Enabled = false;
            ProgJobLabel->Enabled    = false;
            ProgDateLabel->Enabled   = false;
            ProgTimeLabel->Enabled   = false;
            ProgLeftLabel->Enabled   = false;
            ProgVolLabel1->Enabled   = false;
            ProgVolLabel2->Enabled   = false;
            ProgTotalLabel2->Enabled = false;
        }
    }
    else
    {
        // Disable job operations
        StartItem->Enabled = false;
        DelItem->Enabled   = false;
        UpItem->Enabled    = false;
        DownItem->Enabled  = false;

        // Reset progress info
        ResetProgInfo();
    }

    JobListSelectItemUpdateOnly = false;
}

/*---------------------------------------------------------------------------*/

void __fastcall TFSKMenuForm::NewItemClick(TObject *Sender)
{
    FSKShowNewForm();
    //Hide();
    Enabled = false;
}

void __fastcall TFSKMenuForm::AddItemClick(TObject *Sender)
{
    char Path[MAX_PATH];

    // Open folder broswer dialog
    if (BrowseForFolder(Handle, "Select folder containing the job.", NULL, NULL, true, Path))
    {
        if (Path[0] == '\\' && Path[1] == '\\')
            Application->MessageBoxA("Jobs cannot be processed from a remote computer!",
                                     FSK_PRODUCT_NAME, MB_ICONEXCLAMATION | MB_OK);
        else
        {
            bool Exists;

            if (AddJob(AnsiString(Path), Exists))
            {
                if (Exists)
                    Application->MessageBoxA("Job is already added to the list.",
                                     FSK_PRODUCT_NAME, MB_ICONINFORMATION | MB_OK);
            }
            else
                Application->MessageBoxA("The selected folder does not contain a valid job!",
                                     FSK_PRODUCT_NAME, MB_ICONERROR | MB_OK);
        }
    }
}

void __fastcall TFSKMenuForm::DelItemClick(TObject *Sender)
{
    AnsiString Key, NewKey; // registry key name

    if (JobList->Selected->StateIndex == JS_Pending &&
        Application->MessageBoxA("Do you really want remove this job from the queue?", FSK_PRODUCT_NAME, MB_YESNO | MB_ICONASTERISK) == mrNo)
        return;

    // Adjust batch start button state
    if (JobList->Selected->StateIndex == JS_Pending)
    {
        int VolDone, VolTotal;

        PendingJobCount--;
        ShutJobUpDown->Max--;
        sscanf(JobList->Selected->SubItems->Strings[JI_Vol].c_str(), "%d / %d", &VolDone, &VolTotal);
        ShutVolUpDown->Max -= VolTotal - VolDone;

        UpdateShutValues();
    }

    StartAllItem->Enabled = PendingJobCount > 0;

    // Remove from listview
    JobList->DeleteSelected();

    // Remove from registry
    SaveJobList();
}

void __fastcall TFSKMenuForm::UpItemClick(TObject *Sender)
{
    int Index;

    Index = JobList->Selected->Index;
    SwapJobs(Index, Index - 1);
    JobList->ItemIndex = Index - 1;

    if (JobList->Items->Item[Index]->Focused)
    {
        JobList->Items->Item[Index]->Focused = false;
        JobList->Items->Item[Index - 1]->Focused = true;
    }
}


void __fastcall TFSKMenuForm::DownItemClick(TObject *Sender)
{
    int Index;

    Index = JobList->Selected->Index;
    SwapJobs(Index, Index + 1);
    JobList->ItemIndex = Index + 1;

    if (JobList->Items->Item[Index]->Focused)
    {
        JobList->Items->Item[Index]->Focused = false;
        JobList->Items->Item[Index + 1]->Focused = true;
    }
}

void __fastcall TFSKMenuForm::StartItemClick(TObject *Sender)
{
    int VolDone, VolTotal;

    // Work in progress
    IsWorking = true;

    // Disable menu items
    StartItem->Enabled    = false;
    StartAllItem->Enabled = false;
    AbortItem->Enabled    = true;

    NewItem->Enabled      = false;
    AddItem->Enabled      = false;
    DelItem->Enabled      = false;
    UpItem->Enabled       = false;
    DownItem->Enabled     = false;

    // Enable progress info
    ProgLabel1->Enabled     = true;
    ProgLabel2->Enabled     = true;
    ProgLabel3->Enabled     = true;
    ProgLabel4->Enabled     = true;
    ProgTotalLabel1->Enabled     = true;
    ProgJobLabel->Enabled   = true;
    ProgDateLabel->Enabled  = true;
    ProgTimeLabel->Enabled  = true;
    ProgLeftLabel->Enabled  = true;
    ProgVolLabel1->Enabled  = true;
    ProgVolLabel2->Enabled  = true;
    ProgTotalLabel2->Enabled = true;

    // Start job
    IsBatch    = false;
    ShutJobMax = ShutJobUpDown->Max;
    ShutVolMax = ShutVolUpDown->Max;

    sscanf(JobList->Selected->SubItems->Strings[JI_Vol].c_str(), "%d / %d", &VolDone, &VolTotal);
    ShutJobUpDown->Max = 1;
    ShutVolUpDown->Max = VolTotal - VolDone;
    UpdateShutValues();

    StartJob(JobList->ItemIndex);
}

void __fastcall TFSKMenuForm::StartAllItemClick(TObject *Sender)
{
    // Work in progress
    IsWorking = true;

    // Disable menu items
    StartItem->Enabled    = false;
    StartAllItem->Enabled = false;
    AbortItem->Enabled    = true;

    NewItem->Enabled      = false;
    AddItem->Enabled      = false;
    DelItem->Enabled      = false;
    UpItem->Enabled       = false;
    DownItem->Enabled     = false;

    // Enable progress info
    ProgLabel1->Enabled     = true;
    ProgLabel2->Enabled     = true;
    ProgLabel3->Enabled     = true;
    ProgLabel4->Enabled     = true;
    ProgTotalLabel1->Enabled     = true;
    ProgJobLabel->Enabled   = true;
    ProgDateLabel->Enabled  = true;
    ProgTimeLabel->Enabled  = true;
    ProgLeftLabel->Enabled  = true;
    ProgVolLabel1->Enabled  = true;
    ProgVolLabel2->Enabled  = true;
    ProgTotalLabel2->Enabled = true;

    // Start the first pending job
    ShutJobMax = ShutJobUpDown->Max;
    ShutVolMax = ShutVolUpDown->Max;

    for (int i = 0; i < JobList->Items->Count; i++)
        if (JobList->Items->Item[i]->StateIndex == JS_Pending)
        {
            IsBatch = true;
            StartJob(i);
            break;
        }
}

void __fastcall TFSKMenuForm::AbortItemClick(TObject *Sender)
{
    if (Application->MessageBoxA("Do you really want to abort the transcoding process?", FSK_PRODUCT_NAME, MB_YESNO | MB_ICONEXCLAMATION) == mrYes)
        AbortJob();
}

/*---------------------------------------------------------------------------*/

void __fastcall TFSKMenuForm::HelpAboutItemClick(TObject *Sender)
{
    FSKSplashForm->Show();
    FSKSplashForm->FSKAbout(this);
}

void __fastcall TFSKMenuForm::HelpReadmeItemClick(TObject *Sender)
{
    if ((int)ShellExecute(Handle, "open", (FSKAppPath + FSK_README_PATH).c_str(), NULL, NULL, SW_SHOW) <= 32)
        Application->MessageBoxA((AnsiString("Unable to open file \"") + FSK_README_PATH + "\"!").c_str(),
                                 FSK_PRODUCT_NAME, MB_OK | MB_ICONERROR);
}

void __fastcall TFSKMenuForm::HelpLicenseItemClick(TObject *Sender)
{
    if ((int)ShellExecute(Handle, "open", (FSKAppPath + FSK_LICENSE_PATH).c_str(), NULL, NULL, SW_SHOW) <= 32)
        Application->MessageBoxA((AnsiString("Unable to open file \"") + FSK_LICENSE_PATH + "\"!").c_str(),
                                 FSK_PRODUCT_NAME, MB_OK | MB_ICONERROR);
}

/*---------------------------------------------------------------------------*/

void __fastcall TFSKMenuForm::TimerTimer(TObject *Sender)
{
    AnsiString TimeText, TimeLeftText;
    int        TimeRemaining,
               Hours, Minutes;
    double     RefTotalProgress, CurTotalProgress;
    int        RefTimeElapsed, CurTimeElapsed;
    bool       Update = false;

    // Adjust elapsed time
    TimeElapsed += Timer->Interval / 1000;

    Hours   = TimeElapsed / 3600;
    Minutes = (TimeElapsed / 60) - (Hours * 60);
    TimeText.printf("%.2d:%.2d", Hours, Minutes);

    // Estimate remaining time
    RefTotalProgress = Encoder->RefTotalProgress;
    RefTimeElapsed   = Encoder->RefTimeElapsed;

    if (RefTotalProgress > 0.01 && RefTimeElapsed >= 1)
    {
        CurTimeElapsed   = TimeOffset + SecondsBetween(Now(), StartTime);
        CurTotalProgress = (CurTimeElapsed * RefTotalProgress) / RefTimeElapsed;

        if (CurTotalProgress > 0.01)
        {
            TimeRemaining = int((CurTimeElapsed * (100 - CurTotalProgress)) / CurTotalProgress);

            Hours   = TimeRemaining / 3600;
            Minutes = int(floor((double(TimeRemaining) / 60) - (double(Hours) * 60)));
            TimeLeftText.printf("%.2d:%.2d", Hours, Minutes);
            Update = true;
        }
    }

    // Update indicators
    CurJobItem->SubItems->Strings[JI_Time] = TimeText;
    ProgTimeLabel->Caption = TimeText;

    if (Update)
    {
        CurJobItem->SubItems->Strings[JI_TimeLeft] = TimeLeftText;
        ProgLeftLabel->Caption = TimeLeftText;
    }
}

/*---------------------------------------------------------------------------*/

void __fastcall TFSKMenuForm::ShutCheckClick(TObject *Sender)
{
    if (ShutCheck->Checked && ShutJobEdit->Text.ToInt() && ShutVolEdit->Text.ToInt())
    {
        // Enable controls
        ShutAllRadio->Enabled  = true;

        ShutJobRadio->Enabled  = true;
        if (ShutJobRadio->Checked)
        {
            ShutJobEdit->Enabled = true;
            if (ShutJobUpDown->Max == 1)
                ShutJobUpDown->Enabled = false;
            else
                ShutJobUpDown->Enabled = true;
        }
        else
        {
            ShutJobEdit->Enabled   = false;
            ShutJobUpDown->Enabled = false;
        }

        ShutVolRadio->Enabled  = true;
        if (ShutVolRadio->Checked)
        {
            ShutVolEdit->Enabled = true;
            if (ShutVolUpDown->Max == 1)
                ShutVolUpDown->Enabled = false;
            else
                ShutVolUpDown->Enabled = true;
        }
        else
        {
            ShutVolEdit->Enabled   = false;
            ShutVolUpDown->Enabled = false;
        }
    }
    else
    {
        // Disable controls
        ShutAllRadio->Enabled  = false;
        ShutJobRadio->Enabled  = false;
        ShutJobEdit->Enabled   = false;
        ShutJobUpDown->Enabled = false;
        ShutVolRadio->Enabled  = false;
        ShutVolEdit->Enabled   = false;
        ShutVolUpDown->Enabled = false;
    }
}

void __fastcall TFSKMenuForm::ShutAllRadioClick(TObject *Sender)
{
    // Configure controls
    ShutJobEdit->Enabled   = false;
    ShutJobUpDown->Enabled = false;
    ShutVolEdit->Enabled   = false;
    ShutVolUpDown->Enabled = false;
}

void __fastcall TFSKMenuForm::ShutJobRadioClick(TObject *Sender)
{
    // Configure controls
    ShutJobEdit->Enabled   = true;
    ShutJobUpDown->Enabled = true;

    ShutVolEdit->Enabled   = false;
    ShutVolUpDown->Enabled = false;
}

void __fastcall TFSKMenuForm::ShutVolRadioClick(TObject *Sender)
{
    // Configure controls
    ShutJobEdit->Enabled   = false;
    ShutJobUpDown->Enabled = false;

    ShutVolEdit->Enabled   = true;
    ShutVolUpDown->Enabled = true;
}

/*---------------------------------------------------------------------------*/

void __fastcall TFSKMenuForm::PriorityComboChange(TObject *Sender)
{
    if (PriorityCombo->ItemIndex > 1 && PriorityCombo->ItemIndex > PrevPriority)
    {
        if (Application->MessageBoxA("Setting the priority of the encoder above normal may cause\nthe system to freeze! Are you sure you want to do this?", FSK_PRODUCT_NAME, MB_YESNO | MB_ICONEXCLAMATION) != mrYes)
        {
            PriorityCombo->ItemIndex = PrevPriority;
            return;
        }
    }

    if (Encoder->IsChildRunning())
        Encoder->SetPriorityClass(GetPriority());

    PrevPriority = PriorityCombo->ItemIndex;
}

/*-----------------------------------------------------------------------------
    TFSKMenuForm >> User methods
-----------------------------------------------------------------------------*/

// Load job list from registry
void TFSKMenuForm::LoadJobList()
{
    int         i;                              // counter
    int         Count;                          // number of jobs
    AnsiString  Key;                            // registry key name
    TListItem  *Item;                           // listview item object
    AnsiString  JobPath, _JobPath, JobName, JobDate, JobVol, JobTime, JobTimeLeft; // job info
    int         JobDone;
    int         JobTimeH, JobTimeM,             // job time elapsed
                JobTimeLeftH, JobTimeLeftM;     // job time left
    bool        IsValid;                        // is the job valid?
    FILE       *File;                           // file handle
    char        Buffer[256];                    // character buffer
    int         VolDone, VolTotal;              // number of volumes done / total number of volumes

    // Get joblist
    PendingJobCount    = 0;
    ShutJobUpDown->Max = 0;
    ShutVolUpDown->Max = 0;

    JobList->Clear();

    g_Reg->OpenKey(FSK_JOBS_REGKEY, false);
    Count = g_Reg->ReadInteger(FSK_JOBS_COUNT_REGVAL);
    g_Reg->CloseKey();

    for (i = 0; i < Count; i++)
    {
        Key = AnsiString(FSK_JOBS_REGKEY) + "\\";
        Key.cat_printf("%d", i);
        g_Reg->OpenKey(Key, false);

        // Init job info
        JobVol      = "--";
        JobTime     = "--";
        JobTimeLeft = "--";

        // Get job path
        JobPath = g_Reg->ReadString(FSK_JOB_PATH_REGVAL);
        while (JobPath.Length() && JobPath[JobPath.Length()] == '\\')
            JobPath.SetLength(JobPath.Length() - 1);
        _JobPath = JobPath;
        JobPath = JobPath + FSKJ_INFO_FILENAME;

        // Get job name
        JobName = g_Reg->ReadString(FSK_JOB_NAME_REGVAL);

        // Get job date
        JobDate = g_Reg->ReadString(FSK_JOB_DATE_REGVAL);

        // Get job done flag
        JobDone = g_Reg->ReadInteger(FSK_JOB_DONE_REGVAL);

        // Open job info file
        IsValid = false;

        if (File = fopen(JobPath.c_str(), "rt"))
        {
            // Read stuff
            if ((fscanf(File, "%[^\n]\n\n", Buffer) > 0) && !strcmp(Buffer, FSKJ_ID))
            {
                if ((fscanf(File, "%[^\n]\n", Buffer) > 0) && !strcmp(Buffer, JobName.c_str()))
                {
                    if ((fscanf(File, "%[^\n]\n", Buffer) > 0) && !strcmp(Buffer, JobDate.c_str()))
                    {
                        if (fscanf(File, "%[^\n]\n", Buffer) > 0)
                        {
                            JobVol = Buffer;

                            if (fscanf(File, "%[^\n]\n", Buffer) > 0)
                            {
                                JobTime = Buffer;

                                if (fscanf(File, "%[^\n]\n", Buffer) > 0)
                                {
                                    JobTimeLeft = Buffer;

                                    IsValid = true;
                                }
                            }
                        }
                    }
                }
            }

            fclose(File);
        }

        if (!IsValid)
        {
            // Get job name
            JobName = g_Reg->ReadString(FSK_JOB_NAME_REGVAL);
            // Get job date
            JobDate = g_Reg->ReadString(FSK_JOB_DATE_REGVAL);

            // No additional info
            JobVol      = "--";
            JobTime     = "--";
            JobTimeLeft = "--";
        }

        // Add job
        Item = JobList->Items->Add();

        if (IsValid)
        {
            // Check if it's done
            if ((sscanf(JobVol.c_str(), "%d / %d", &VolDone, &VolTotal) != 2) ||
                (VolDone < 0) || (VolTotal <= 0) || (VolDone > VolTotal) ||
                (JobTime != "--" && ((sscanf(JobTime.c_str(), "%d:%d", &JobTimeH, &JobTimeM) != 2) ||
                (JobTimeH < 0) || (JobTimeM < 0) || (JobTimeM >= 60))) ||
                (JobTimeLeft != "--" && ((sscanf(JobTimeLeft.c_str(), "%d:%d", &JobTimeLeftH, &JobTimeLeftM) != 2) ||
                (JobTimeLeftH < 0) || (JobTimeLeftM < 0) || (JobTimeLeftM >= 60))))
            {
                IsValid = false;

                // Get job name
                JobName = g_Reg->ReadString(FSK_JOB_NAME_REGVAL);
                // Get job date
                JobDate = g_Reg->ReadString(FSK_JOB_DATE_REGVAL);

                // No additional info
                JobVol      = "--";
                JobTime     = "--";
                JobTimeLeft = "--";
            }
            else
            {
                if (VolDone == VolTotal)
                {
                    Item->StateIndex = JS_Done;      // done
                    JobTimeLeft = "--";
                }
                else
                {
                    Item->StateIndex = JS_Pending;   // waiting
                    PendingJobCount++;

                    if (!VolDone)
                    {
                        JobTime     = "--";     // new
                        JobTimeLeft = "--";     // new
                    }

                    ShutJobUpDown->Max++;
                    ShutVolUpDown->Max += VolTotal - VolDone;
                }

                // Reformat texts
                JobVol.printf("%d / %d", VolDone, VolTotal);

                if (JobTime != "--")
                    JobTime.printf("%.2d:%.2d", JobTimeH, JobTimeM);

                if (JobTimeLeft != "--")
                    JobTimeLeft.printf("%.2d:%.2d", JobTimeLeftH, JobTimeLeftM);
            }
        }


        if (!IsValid)
        {
            if (JobDone)
                Item->StateIndex = JS_Done;          // invalid because it's done
            else
                Item->StateIndex = JS_Invalid;       // invalid
        }

        Item->SubItems->Add(JobName);
        Item->SubItems->Add(JobDate);
        Item->SubItems->Add(JobVol);
        Item->SubItems->Add(JobTime);
        Item->SubItems->Add(JobTimeLeft);
        Item->SubItems->Add(_JobPath);

        g_Reg->CloseKey();
    }

    UpdateShutValues();
}

// Save job list to registry
void TFSKMenuForm::SaveJobList()
{
    AnsiString Key; // key name

    // Remove current job list from registry
    if (g_Reg->KeyExists(FSK_JOBS_REGKEY))
    {
        if (!DeleteKeyTree(FSK_JOBS_REGKEY))
            goto Error;
    }

    // Write job count
    if (!g_Reg->OpenKey(FSK_JOBS_REGKEY, true))
        goto Error;

    g_Reg->WriteInteger(FSK_JOBS_COUNT_REGVAL, JobList->Items->Count);
    g_Reg->CloseKey();

    // Write job entries
    for (int i = 0; i < JobList->Items->Count; i++)
    {
        // Create key
        Key = FSK_JOBS_REGKEY;
        Key.cat_printf("\\%d", i);

        if (!g_Reg->OpenKey(Key, true))
            goto Error;

        // Write values
        g_Reg->WriteString(FSK_JOB_NAME_REGVAL, JobList->Items->Item[i]->SubItems->Strings[JI_Name]);
        g_Reg->WriteString(FSK_JOB_DATE_REGVAL, JobList->Items->Item[i]->SubItems->Strings[JI_Date]);
        g_Reg->WriteString(FSK_JOB_PATH_REGVAL, JobList->Items->Item[i]->SubItems->Strings[JI_Path]);
        g_Reg->WriteInteger(FSK_JOB_DONE_REGVAL, JobList->Items->Item[i]->StateIndex == JS_Done);

        // Close key
        g_Reg->CloseKey();
    }

    // Success
    return;

    // Error
Error:
    Application->MessageBoxA("Error updating configuration!", FSK_PRODUCT_NAME, MB_ICONERROR | MB_OK);
    Application->Terminate();
    return;
}

// Add a new job to the list
bool TFSKMenuForm::AddJob(AnsiString Path, bool &Exists)
{
    int         i;                              // counter
    FILE       *File;                           // file handle
    char        Buffer[256];                    // character buffer
    AnsiString  JobPath, JobName, JobDate, JobVol, JobTime, JobTimeLeft; // job info
    int         JobTimeH, JobTimeM,             // job time elapsed
                JobTimeLeftH, JobTimeLeftM;     // job time left
    bool        IsValid;                        // is the job valid?
    int         JobIndex;                       // index of the job in the queue
    AnsiString  Key;                            // registry key name
    TListItem  *Item;                           // listview item object pointer
    int         VolDone, VolTotal;              // number of volumes done / total number of volumes
    AnsiString  VolPath;                        // path of the current volume

    JobPath = Path + FSKJ_INFO_FILENAME;

    // Open job info file
    IsValid = false;

    if (File = fopen(JobPath.c_str(), "rt"))
    {
        // Read stuff
        if ((fscanf(File, "%[^\n]\n\n", Buffer) > 0) && !strcmp(Buffer, FSKJ_ID))
        {
            if ((fscanf(File, "%[^\n]\n", Buffer) > 0))
            {
                JobName = Buffer;

                if ((fscanf(File, "%[^\n]\n", Buffer) > 0))
                {
                    JobDate = Buffer;

                    if (fscanf(File, "%[^\n]\n", Buffer) > 0)
                    {
                        JobVol = Buffer;

                        if (fscanf(File, "%[^\n]\n", Buffer) > 0)
                        {
                            JobTime = Buffer;

                            if (fscanf(File, "%[^\n]\n", Buffer) > 0)
                            {
                                JobTimeLeft = Buffer;

                                IsValid = true;
                            }
                        }
                    }
                }
            }
        }

        fclose(File);
    }

    if (!IsValid)
        return false;

    // Check if it already exists in the list
    Exists = false;

    for (i = 0; i < JobList->Items->Count; i++)
    {
        Item = JobList->Items->Item[i];
        if (Item->SubItems->Strings[JI_Name] == JobName && Item->SubItems->Strings[JI_Date] == JobDate)
        {
            Exists = true;
            return true;
        }
    }

    // Check for output job files
    if ((sscanf(JobVol.c_str(), "%d / %d", &VolDone, &VolTotal) != 2) ||
        (VolDone < 0) || (VolTotal <= 0) || (VolDone > VolTotal) ||
        (JobTime != "--" && ((sscanf(JobTime.c_str(), "%d:%d", &JobTimeH, &JobTimeM) != 2) ||
        (JobTimeH < 0) || (JobTimeM < 0) || (JobTimeM >= 60))) ||
        (JobTimeLeft != "--" && ((sscanf(JobTimeLeft.c_str(), "%d:%d", &JobTimeLeftH, &JobTimeLeftM) != 2) ||
        (JobTimeLeftH < 0) || (JobTimeLeftM < 0) || (JobTimeLeftM >= 60))))
        return false;

    for (i = 1; i <= VolTotal; i++)
    {
        VolPath = Path + "\\.fskj\\";
        VolPath.cat_printf(FSK_OUT_FILENAME, i);
        VolPath.cat_printf(".rpjf");

        if (!FileExists(VolPath))
            return false;
    }

    // Add job to the list
    Item = JobList->Items->Add();

    // Check if it's done
    if (VolDone == VolTotal)
    {
        Item->StateIndex = JS_Done;      // done
        JobTimeLeft = "--";
    }
    else
    {
        Item->StateIndex = JS_Pending;   // waiting
        PendingJobCount++;

        if (!VolDone)
        {
            JobTime     = "--";     // new
            JobTimeLeft = "--";     // new
        }

        ShutJobUpDown->Max++;
        ShutVolUpDown->Max += VolTotal - VolDone;
        UpdateShutValues();
    }

    // Reformat texts
    JobVol.printf("%d / %d", VolDone, VolTotal);

    if (JobTime != "--")
        JobTime.printf("%.2d:%.2d", JobTimeH, JobTimeM);

    if (JobTimeLeft != "--")
        JobTimeLeft.printf("%.2d:%.2d", JobTimeLeftH, JobTimeLeftM);

    Item->SubItems->Add(JobName);
    Item->SubItems->Add(JobDate);
    Item->SubItems->Add(JobVol);
    Item->SubItems->Add(JobTime);
    Item->SubItems->Add(JobTimeLeft);
    Item->SubItems->Add(Path);

    // Save to registry
    SaveJobList();

    // Set batch start button
    StartAllItem->Enabled = PendingJobCount > 0;

    // Success
    return true;
}

// Swap two jobs in the list
void TFSKMenuForm::SwapJobs(int i, int j)
{
    TListItem  *Item;               // listview item object pointer
    AnsiString  Key1, Key2, KeyT;   // registry key names

    // Swap in listview
    JobList->Items->BeginUpdate();
    Item = new TListItem(JobList->Items);

    Item->Assign(JobList->Items->Item[i]);
    JobList->Items->Item[i]->Assign(JobList->Items->Item[j]);
    JobList->Items->Item[j]->Assign(Item);

    Item->Free();
    JobList->Items->EndUpdate();

    // Swap in registry
    SaveJobList();
}

// Write job info file
bool TFSKMenuForm::WriteJobInfo(AnsiString Filename, TListItem *Item)
{
    FILE *File;   // file handle

    if ((File = fopen(Filename.c_str(), "wt")) == NULL)
        return false;

    fprintf(File, FSKJ_ID);
    fprintf(File, "\n\n%s\n%s\n%s\n%s\n%s",
                  Item->SubItems->Strings[JI_Name],
                  Item->SubItems->Strings[JI_Date],
                  Item->SubItems->Strings[JI_Vol],
                  Item->SubItems->Strings[JI_Time],
                  Item->SubItems->Strings[JI_TimeLeft]);

    fclose(File);
    return true;
}

/*---------------------------------------------------------------------------*/

// Start encoding a job
void TFSKMenuForm::StartJob(int JobIndex)
{
    AnsiString Path;               // job path (.fskj)
    int        VolDone, VolTotal;  // number of volumes done / total number of volumes

    // Select job
    CurJobItem = JobList->Items->Item[JobIndex];
    IsWorking = false;
    JobListSelectItemUpdateOnly = true;
    JobListSelectItem(NULL, CurJobItem, true);
    IsWorking = true;

    // Change job status
    CurJobItem->StateIndex = JS_Working;

    // Get path
    Path = CurJobItem->SubItems->Strings[JI_Path];
    Path += FSKJ_PATH;

    // Get info about job
    sscanf(CurJobItem->SubItems->Strings[JI_Vol].c_str(), "%d / %d", &VolDone, &VolTotal);
    LastTime = CurJobItem->SubItems->Strings[JI_Time];
    LastTimeLeft = CurJobItem->SubItems->Strings[JI_TimeLeft];

    // Change to the job's internal directory
    SetCurrentDir(Path);

    // Start time
    StartTimer();

    // Start volume
    StartJobVol(VolDone + 1, VolTotal);
}

// Abort encoding a job
void TFSKMenuForm::AbortJob()
{
    Screen->Cursor = crHourGlass;

    // Wait
    while (IsWorking && !Encoder->IsWorking())
        Sleep(50);

    // Stop
    if (IsWorking && Encoder->IsWorking())
        Encoder->Stop(false);

    Screen->Cursor = crDefault;
}

/*---------------------------------------------------------------------------*/

// Start encoding a job's volume
void TFSKMenuForm::StartJobVol(int Vol, int VolTotal)
{
    AnsiString Text, RPJFName;    // name of .rpjf file

    // Update indicators
    Text.printf("Volume %d / %d:", Vol, VolTotal);
    ProgVolLabel1->Caption = Text;

    ProgVolBar->Position   = 0;
    ProgVolLabel2->Caption = "0.0 %";
    ProgVolLabel2->Refresh();

    // Get producer job name
    RPJFName.printf(FSK_OUT_FILENAME".rpjf", Vol);

    // Start
    Encoder->TerminateChildProcess();
    if (!Encoder->Start(FSKEncoderPath, RPJFName, Vol, VolTotal, GetPriority()))
        _AbortJobVol(true, "Unable to initialize encoder!");
}

// Finished encoding a job's volume
void TFSKMenuForm::_EndJobVol(int Vol, int VolTotal)
{
    bool       Done, ShutFlag;
    AnsiString Text, FilePath;

    // Update job progress
    Text.printf("%d / %d", Vol, VolTotal);
    CurJobItem->SubItems->Strings[JI_Vol] = Text;

    if (Vol == VolTotal)
    {
        CurJobItem->SubItems->Strings[JI_TimeLeft] = "--";
        ProgLeftLabel->Caption = "--";
    }

    ShutFlag = ShutCheck->Checked && ShutVolRadio->Checked && (ShutVolUpDown->Position == 1);
    if (ShutCheck->Checked && ShutVolRadio->Checked)
        ShutVolUpDown->Position--;
    ShutVolUpDown->Max--;
    ShutVolMax--;

    // Update job info file
    FilePath = CurJobItem->SubItems->Strings[JI_Path];
    FilePath += FSKJ_INFO_FILENAME;
    WriteJobInfo(FilePath, CurJobItem);

    LastTime = CurJobItem->SubItems->Strings[JI_Time];
    LastTimeLeft = CurJobItem->SubItems->Strings[JI_TimeLeft];

    // Check if the job is done
    if (Vol == VolTotal)
    {
        // Change job status
        CurJobItem->StateIndex = JS_Done;

        // Remove job files
        SetCurrentDir(CurJobItem->SubItems->Strings[JI_Path]);
        RemoveDirTree(CurJobItem->SubItems->Strings[JI_Path] + FSKJ_PATH);

        // Update job list
        SaveJobList();
    }

    // Check if we must turn off the computer
    if (ShutFlag)
    {
        // Change job status
        if (Vol != VolTotal)
        {
            Application->Title = FSKAppTitle;
            FSKMenuForm->Caption = FSKMenuForm->FormCaption;
            CurJobItem->StateIndex = JS_Pending;
        }

        // Stop timer
        StopTimer();

        UpdateShutValues();

        // Beep
        MessageBeep(MB_OK);

        // Enable menu items
        StartItem->Enabled    = true;
        StartAllItem->Enabled = true;
        AbortItem->Enabled    = false;

        NewItem->Enabled      = true;
        AddItem->Enabled      = true;

        // Done
        IsWorking = false;
        ShutVolUpDown->Max = ShutVolMax;
        ShutJobUpDown->Max = ShutJobMax;
        UpdateShutValues();

        // Select job (if required)
        if (JobList->ItemIndex < 0)
            JobListSelectItem(NULL, NULL, false);
        else
            JobListSelectItem(NULL, JobList->Items->Item[JobList->ItemIndex], true);

        // Turn off computer
        Refresh();
        Shutdown();
        return;
    }

    // Check if the job is done
    if (Vol == VolTotal)
    {
        // Stop timer
        StopTimer();

        // Change job status
        PendingJobCount--;

        ShutFlag = ShutCheck->Checked && ShutJobRadio->Checked && (ShutJobUpDown->Position == 1);
        if (ShutCheck->Checked && ShutJobRadio->Checked)
            ShutJobUpDown->Position--;
        ShutJobUpDown->Max--;
        ShutJobMax--;
        UpdateShutValues();

        ProgTotalBar->Position  = 100;
        ProgTotalLabel2->Caption = "100.0 %";

        // Beep
        MessageBeep(MB_OK);

        Done = true;
        if (IsBatch && !ShutFlag)
        {
            // Start the next pending job
            for (int i = CurJobItem->Index + 1; i < JobList->Items->Count; i++)
                if (JobList->Items->Item[i]->StateIndex == JS_Pending)
                {
                    Done = false;
                    StartJob(i);
                    break;
                }
        }

        if (Done)
        {
            // Enable menu items
            StartItem->Enabled    = true;
            StartAllItem->Enabled = true;
            AbortItem->Enabled    = false;

            NewItem->Enabled      = true;
            AddItem->Enabled      = true;

            // Done
            IsWorking = false;
            ShutVolUpDown->Max = ShutVolMax;
            ShutJobUpDown->Max = ShutJobMax;
            UpdateShutValues();

            // Select job (if required)
            if (JobList->ItemIndex < 0)
                JobListSelectItem(NULL, NULL, false);
            else
                JobListSelectItem(NULL, JobList->Items->Item[JobList->ItemIndex], true);

            // Check if we must turn off the computer
            ShutFlag |= ShutCheck->Checked && ShutAllRadio->Checked;
            if (ShutFlag)
            {
                Refresh();
                Shutdown();
                return;
            }
        }
    }
    else
    {
        UpdateShutValues();
        Refresh();

        // Next volume
        StartJobVol(Vol + 1, VolTotal);
    }

    Refresh();
}

// Abort
void TFSKMenuForm::_AbortJobVol(bool Error, AnsiString Message)
{
    bool Done, ShutFlag = false;
    int  VolDone, VolTotal;

    // Stop timer
    StopTimer();

    if (Error)
    {
        // Error
        if (IsBatch || ShutCheck->Checked)
            MessageBeep(MB_ICONERROR);
        else
            Application->MessageBoxA(Message.c_str(), FSK_PRODUCT_NAME, MB_ICONERROR);

        CurJobItem->StateIndex = JS_Pending;

        // Check if we must turn off the computer
        sscanf(CurJobItem->SubItems->Strings[JI_Vol].c_str(), "%d / %d", &VolDone, &VolTotal);

        if (ShutCheck->Checked)
        {
            if (ShutJobRadio->Checked)
            {
                if (ShutJobUpDown->Position == 1)
                    ShutFlag = true;
                else
                    ShutJobUpDown->Position--;
            }
            else
                if (ShutVolRadio->Checked)
                {
                    if (ShutVolUpDown->Position <= VolTotal - VolDone)
                        ShutFlag = true;
                    else
                        ShutVolUpDown->Position -= VolTotal - VolDone;
                }
                else
                    ShutFlag = false;
        }
        else
            ShutFlag = false;

        ShutVolUpDown->Max -= VolTotal - VolDone;
        ShutJobUpDown->Max--;
        UpdateShutValues();
    }
    else
    {
        // User abort
        CurJobItem->StateIndex = JS_Pending;

        // Beep
        MessageBeep(MB_ICONEXCLAMATION);
    }

    Done = true;
    if (Error && IsBatch && !ShutFlag)
    {
        // Start the next pending job
        for (int i = CurJobItem->Index + 1; i < JobList->Items->Count; i++)
            if (JobList->Items->Item[i]->StateIndex == JS_Pending)
            {
                Done = false;
                StartJob(i);
                break;
            }
    }

    if (!Done)
    {
        Refresh();
        return;
    }

    // Reset timestamps
    CurJobItem->SubItems->Strings[JI_Time]     = LastTime;
    CurJobItem->SubItems->Strings[JI_TimeLeft] = LastTimeLeft;

    ProgVolBar->Position   = 0;
    ProgVolLabel2->Caption = "0.0 %";

    // Enable menu items
    StartItem->Enabled    = true;
    StartAllItem->Enabled = true;
    AbortItem->Enabled    = false;

    NewItem->Enabled      = true;
    AddItem->Enabled      = true;

    // Done
    IsWorking = false;
    ShutVolUpDown->Max = ShutVolMax;
    ShutJobUpDown->Max = ShutJobMax;
    UpdateShutValues();

    // Select job (if required)
    if (JobList->ItemIndex < 0)
        JobListSelectItem(NULL, NULL, false);
    else
        JobListSelectItem(NULL, JobList->Items->Item[JobList->ItemIndex], true);

    // Check if we must turn off the computer
    ShutFlag |= ShutCheck->Checked && ShutAllRadio->Checked && Error && Done;
    if (ShutFlag)
    {
        Refresh();
        Shutdown();
        return;
    }

    Refresh();
}

/*---------------------------------------------------------------------------*/

// Start timer
void TFSKMenuForm::StartTimer()
{
    AnsiString Text;

    // Set elapsed time
    Text = CurJobItem->SubItems->Strings[JI_Time];
    if (Text == "--")
    {
        TimeElapsed = TimeOffset = 0;

        CurJobItem->SubItems->Strings[JI_Time] = "00:00";
        ProgTimeLabel->Caption = "00:00";
    }
    else
    {
        int Hours, Minutes;

        sscanf(Text.c_str(), "%d:%d", &Hours, &Minutes);
        TimeElapsed = Minutes * 60 + Hours * 3600;
        TimeOffset  = TimeElapsed;

        ProgTimeLabel->Caption = Text;
    }

    // Update indicators
    ProgLeftLabel->Caption = CurJobItem->SubItems->Strings[JI_TimeLeft];

    // Get current time
    StartTime = Now();
    Encoder->RefTimeElapsed   = -1;
    Encoder->RefTotalProgress = -1;

    // Enable timer
    Timer->Enabled = true;
}

// Stop timer
void TFSKMenuForm::StopTimer()
{
    // Disable timer
    Timer->Enabled = false;
}

/*---------------------------------------------------------------------------*/

// Reset progress info
void TFSKMenuForm::ResetProgInfo()
{
    ProgJobLabel->Caption   = "--";
    ProgDateLabel->Caption  = "--";
    ProgTimeLabel->Caption  = "--";
    ProgLeftLabel->Caption  = "--";
    ProgVolLabel1->Caption  = "Volume:";
    ProgVolLabel2->Caption  = "0.0 %";
    ProgTotalLabel2->Caption = "0.0 %";

    ProgVolBar->Position    = 0;
    ProgTotalBar->Position  = 0;

    ProgLabel1->Enabled     = false;
    ProgLabel2->Enabled     = false;
    ProgLabel3->Enabled     = false;
    ProgLabel4->Enabled     = false;
    ProgTotalLabel1->Enabled     = false;
    ProgJobLabel->Enabled   = false;
    ProgDateLabel->Enabled  = false;
    ProgTimeLabel->Enabled  = false;
    ProgLeftLabel->Enabled  = false;
    ProgVolLabel1->Enabled  = false;
    ProgVolLabel2->Enabled  = false;
    ProgTotalLabel2->Enabled = false;
}

// Update shutdown related values (max number of jobs and volumes)
void TFSKMenuForm::UpdateShutValues()
{
    if (!ShutJobUpDown->Max)
        ShutJobEdit->Text = "0";
    else
    {
        if (ShutJobEdit->Text.ToInt() > ShutJobUpDown->Max)
            ShutJobUpDown->Position = ShutJobUpDown->Max;

        if (!ShutJobEdit->Text.ToInt() && ShutJobUpDown->Max)
            ShutJobEdit->Text = 1;
    }

    if (!ShutVolUpDown->Max)
        ShutVolEdit->Text = "0";
    else
    {
        if (ShutVolEdit->Text.ToInt() > ShutVolUpDown->Max)
            ShutVolUpDown->Position = ShutVolUpDown->Max;

        if (!ShutVolEdit->Text.ToInt() && ShutVolUpDown->Max)
            ShutVolEdit->Text = 1;
    }

    ShutCheckClick(NULL);
}

// Shutdown computer
void TFSKMenuForm::Shutdown()
{
    //Application->MessageBoxA("Shutdown", "debug", MB_OK);
    ::Shutdown(SA_PowerOff);
}

// Get encoder priority from the respective combo box
DWORD TFSKMenuForm::GetPriority()
{
    switch (PriorityCombo->ItemIndex)
    {
    case 0:
        return IDLE_PRIORITY_CLASS;

    case 1:
        return NORMAL_PRIORITY_CLASS;

    case 2:
        return HIGH_PRIORITY_CLASS;

    case 3:
        return REALTIME_PRIORITY_CLASS;

    default:
        return NORMAL_PRIORITY_CLASS;
    }
}

/*---------------------------------------------------------------------------*/

// Registry error, terminate application
void TFSKMenuForm::RegError()
{
    Application->MessageBoxA("Registry entries are corrupted! Please restart FilmShrink.", FSK_PRODUCT_NAME, MB_ICONERROR | MB_OK);
    Application->Terminate();
}

// Delete an entire registry key tree
bool TFSKMenuForm::DeleteKeyTree(AnsiString Key)
{
    TStringList *SubKeys;           // list of subkeys
    bool         Success = true;

    // Open key
    if (!g_Reg->KeyExists(Key) || !g_Reg->OpenKey(Key, false))
        return false;

    // Get subkeys
    SubKeys = new TStringList();
    g_Reg->GetKeyNames(SubKeys);
    g_Reg->CloseKey();

    // Remove subkeys
    for (int i = 0; i < SubKeys->Count; i++)
        if (!DeleteKeyTree(Key + "\\" + SubKeys->Strings[i]))
            Success = false;

    delete SubKeys;

    // Remove key
    if (!g_Reg->DeleteKey(Key))
        Success = false;

    return Success;
}

/*=============================================================================
    TFSKEncoder class
=============================================================================*/

/*-----------------------------------------------------------------------------
    TFSKEncoder >> Control
-----------------------------------------------------------------------------*/

// Start the encoding process
bool TFSKEncoder::Start(AnsiString EncoderPath, AnsiString JobName, int Vol, int VolTotal, DWORD Priority)
{
    AnsiString CmdLine;

    // Configuration
    IsActive  = false;  // currently inactive
    IsDone    = false;
    IsAborted = false;  // no user abort
    IsKilled  = false;  // no selfkill
    PID       = 0;
    CurVol    = Vol;
    NumVol    = VolTotal;

    // Get encoding speed ratio (if 2 pass)
    if (!FSKGetRPJFVideoComplexity(JobName.c_str(), IsTwoPass, EncSpeed, Pass1Speed))
        return false;

    if (EncSpeed < 0)
        EncSpeed = 90;
    EncSpeed = FSKVideoEHQSpeed(EncSpeed);

    if (IsTwoPass)
    {
        if (Pass1Speed < 0)
            Pass1Speed = 50;
        Pass1Speed = FSKVideoEHQSpeed(Pass1Speed);
    }
    else
        Pass1Speed = 0;

    TotalSpeed = EncSpeed + Pass1Speed;

    PrevProgress = PrevPass1Progress = -1;

    // Delete AVS debug file
    if (FileExists(FSKJ_AVS_DEBUG_FILENAME))
    {
        if (FileSetAttr(FSKJ_AVS_DEBUG_FILENAME, 0) || !DeleteFile(FSKJ_AVS_DEBUG_FILENAME))
            return false;
    }

    // Start child process
    CmdLine  = EncoderPath;
    CmdLine += " -j \"";
    CmdLine += JobName;
    CmdLine += "\" -daw -lc \"e,i\" -pid pid";   // disable audio watchdog, show only errors and info, save pid

    return StartChildProcess(CmdLine.c_str(), Priority, false);
}

// Abort the encoding process
void TFSKEncoder::Stop(bool Error)
{
    // Following 2 strings are defined by the producer.exe when it registers for window events
    const char ProducerSignalWndMsg[]    = "Helix Producer signal catcher";
    const char ProducerSignalClassName[] = "Helix Producer signal catcher";

    int   WndID;
    char  WndName[500];
    HWND  hwnd;
    int   res = 0;

    if (Error)
        IsKilled  = true;
    else
        IsAborted = true;

    if (!IsChildRunning() || !IsWorking())
        return;

    // Get the Window ID for the Helix Producer Signal Handler
    WndID = RegisterWindowMessage(ProducerSignalWndMsg);
    if (!WndID)
        goto Kill;

    // Create unique ID for a specific instance producer.exe by combining the PID and Window Message
    sprintf(WndName, "%d %s", PID, ProducerSignalWndMsg);
    hwnd = FindWindow(ProducerSignalClassName, WndName);
    if (!hwnd)
        goto Kill;

    // Cancel execution
    res = PostMessage(hwnd, WndID, 1, NULL);
    return;

    // Kill
Kill:
    TerminateChildProcess();
}

/*-----------------------------------------------------------------------------
    TFSKEncoder >> Events
-----------------------------------------------------------------------------*/

void TFSKEncoder::OnChildStdOutWriteLn(LPCSTR lpszOutput)
{
    const char *p;          // pointer to a string
    int         Progress;   // progress (in %)

    if (IsKilled || IsAborted)
        return;

    // Parse message
    if ((p = strstr(lpszOutput, "Encoding: ")) || (p = strstr(lpszOutput, "Encoding@")))
    {
        char tmp1[256], tmp2[256];   // temporary strings

        Progress = -1;
        sscanf(p, "Encoding%[^:]: %[^(](%d%%)", tmp1, tmp2, &Progress);
        if (Progress >= 0)
        {
            // Update progress
            if (IsTwoPass)
                SetProgress(double((Pass1Speed * 100) + Progress * EncSpeed) / TotalSpeed, PrevProgress != Progress);
            else
                SetProgress(double(Progress), PrevProgress != Progress);

            PrevProgress = Progress;
        }
    }
    else
        if (p = strstr(lpszOutput, "Analysis: "))
        {
            if (sscanf(p, "Analysis: %d%%", &Progress) == 1)
            {
                // Update progress
                SetProgress(double(Progress * Pass1Speed) / TotalSpeed, PrevPass1Progress != Progress);
                PrevPass1Progress = Progress;
            }
        }
        else
            if (p = strstr(lpszOutput, "Error: "))
            {
                // Error
                IsKilled = true;
                return;
            }
            else
                if (p = strstr(lpszOutput, "Info: Starting encode"))
                {
                    // Check AVS debug file
                    if (!FileExists(FSKJ_AVS_DEBUG_FILENAME))
                    {
                        // Error
                        IsKilled = true;
                        TerminateChildProcess();
                        return;
                    }

                    // Update progress
                    SetProgress(0, false);
                }
                else
                    if (p = strstr(lpszOutput, "Done "))
                    {
                        // Set "Done" flag
                        IsDone = true;
                    }
                    else
                        if (p = strstr(lpszOutput, "Helix DNA(TM) Producer 10"))
                        {
                            FILE *PIDFILE;

                            IsActive = true;

                            // Read PID file
                            if ((PIDFILE = fopen("pid", "rt")) == NULL)
                            {
                                IsKilled = true;
                                TerminateChildProcess();
                                return;
                            }

                            if (fscanf(PIDFILE, "%d", &PID) != 1)
                            {
                                fclose(PIDFILE);
                                IsKilled = true;
                                TerminateChildProcess();
                                return;
                            }

                            fclose(PIDFILE);
                        }
}

void TFSKEncoder::OnChildTerminate()
{
    bool _IsActive;

    _IsActive = IsActive;
    IsActive  = false;
    Application->Title = FSKAppTitle;
    FSKMenuForm->Caption = FSKMenuForm->FormCaption;

    if (IsAborted)
        // User abort
        FSKMenuForm->_AbortJobVol(false);
    else
        if (IsKilled)
            // General error
            FSKMenuForm->_AbortJobVol(true, "Encoding failed!");
        else
            if (IsDone)
            {
                // Update progress
                SetProgress(100, true);

                // Success
                FSKMenuForm->_EndJobVol(CurVol, NumVol);
            }
            else
                if (!_IsActive)
                    // Encoder failed to start
                    FSKMenuForm->_AbortJobVol(true, "Unable to initialize encoder!");
                else
                    // Abnormal termination
                    FSKMenuForm->_AbortJobVol(true, "Encoder was abnormally terminated!");
}

/*-----------------------------------------------------------------------------
    TFSKEncoder >> Form control methods
-----------------------------------------------------------------------------*/

// Update the progress indicators
void TFSKEncoder::SetProgress(double Progress, bool Estimate)
{
    AnsiString Text;             // text buffer
    double     TotalProgress;    // total progress
    AnsiString TimeText, TimeLeftText;
    int        TimeRemaining,
               Hours, Minutes;

    // Volume progress
    FSKMenuForm->ProgVolBar->Position = int(Progress);
    Text.printf("%.1f %%", Progress);
    FSKMenuForm->ProgVolLabel2->Caption = Text;

    // Total progress
    TotalProgress = (((CurVol - 1) * 100) + Progress) / NumVol;

    FSKMenuForm->ProgTotalBar->Position = int(TotalProgress);
    Text.printf("%.1f %%", TotalProgress);
    FSKMenuForm->ProgTotalLabel2->Caption = Text;

    if (int(ceil(TotalProgress)) == 100)
    {
        Application->Title = FSKAppTitle;
        FSKMenuForm->Caption = FSKMenuForm->FormCaption;
    }
    else
    {
        Text = FSK_PRODUCT_NAME;
        Text.cat_printf(" - %.1f%%", TotalProgress);
        Application->Title = Text;

        Text = FSKMenuForm->FormCaption;
        Text.cat_printf(" - Encoding %.1f%%", TotalProgress);
        FSKMenuForm->Caption = Text;
    }

    // Estimate remaining time
    if (Estimate)
    {
        RefTotalProgress = TotalProgress;
        RefTimeElapsed   = FSKMenuForm->TimeOffset + SecondsBetween(Now(), FSKMenuForm->StartTime);
    }
}

