/*=============================================================================
    Project Name:
            FilmShrink

	Module Name:
			FSKNewThreads.cpp

	Abstract:
			FilmShrink "New Job" threads

    License:
            GNU General Public License (GPL)

	Revision History:
			* Oct 13, 2004:  Created by Attila Afra
            * Jan  4, 2005:  Subpicture support
            * Jan 30, 2005:  Audio and subpicture detection fixed
            * Feb  7, 2005:  Crop value adjustment
            * Mar  4, 2005:  Serious analysis bugs fixed
            * Mar  5, 2005:  Index import support
            * Mar 24, 2005:  Index import bug workaround

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

// Disable warnings
#pragma warn -8060  // 'Possibly incorrect assignment'

// Includes
#include <stdio.h>
#include <DateUtils.hpp>

#include "DVDSum.h"
#include "m2preview.h"

#include "FSKGlobal.h"
#include "FSKMenu.h"
#include "FSKNewWait.h"
#include "FSKNewThreads.h"

/*=============================================================================
    TFSKAnalyzeThread
=============================================================================*/

/*-----------------------------------------------------------------------------
    TFSKAnalyzeThread >> Constructor
-----------------------------------------------------------------------------*/

// Constructor
__fastcall TFSKAnalyzeThread::TFSKAnalyzeThread(bool CreateSuspended) : TThread(CreateSuspended)
{
    // Selfconfigure
    FreeOnTerminate = false;

    // Initialize values
    Term = false;
    IdxInfo.IsActive = false;
    IdxInfo.IsSuspended = false;
    IdxInfo.IsSuccess = false;
}

// Destructor
__fastcall TFSKAnalyzeThread::~TFSKAnalyzeThread()
{
    //Application->MessageBoxA("Free", "TFSKAnalyzeThread", MB_OK);
}

/*-----------------------------------------------------------------------------
    TFSKAnalyzeThread >> Main methods
-----------------------------------------------------------------------------*/

// Main thread code
void __fastcall TFSKAnalyzeThread::Execute()
{
    int   i;            // counter
    FILE *File;         // file handle
    AnsiString Text;    // temporary string

    // Initialization
    IdxLog = new TStringList;
    IdxLog->Sorted = false;

    // Set progress bar to 0%
    SetProgress(0);

    // Create job
    if (ACreateJob())
        Info("Job created");
    else
    {
        delete IdxLog;
        FatalError("Job creation failed", "Error creating new job!");
        return;
    }

    // Check if we have to import the index from another job
    Import = ACheckImport();

    if (Import)
    {
        // Import index
        if (!AImportStreams())
        {
            delete IdxLog;
            return;
        }
    }
    else
    {
        // Index
        IdxLog->Clear();
        IdxLog->Add(AnsiString(FSKJ_IDXLOG_ID) + "\n\n");
        IdxLog->Add(DVDPath + "\n");
        Text.printf("%.8x\n\n", DVDSum);
        IdxLog->Add(Text);

        // Process all streams
        if (!AProcessStreams())
        {
            delete IdxLog;
            return;
        }

        // Write .idxlog file
        if ((File = fopen((JobPath + FSKJ_IDXLOG_ID).c_str(), "wt")) != NULL)
        {
            for (i = 0; i < IdxLog->Count; i++)
                fputs(IdxLog->Strings[i].c_str(), File);
            fclose(File);
        }
        delete IdxLog;
    }

    // Success
    Finish(true);
}

// Query suspend
void TFSKAnalyzeThread::SuspendQuery()
{
    // Suspend indexing thread, if active
    if (IdxInfo.IsActive)
        m2IndexSuspend(&IdxInfo);

    // Suspend main thread (self)
    Suspend();
}

// Query resume
void TFSKAnalyzeThread::ResumeQuery()
{
    // Resume main thread (self)
    Resume();

    // Resume indexing thread, if active
    if (IdxInfo.IsActive)
        m2IndexResume(&IdxInfo);
}

/*-----------------------------------------------------------------------------
    TFSKAnalyzeThread >> Core
-----------------------------------------------------------------------------*/

// Check if the thread must be terminated and prepare for exit, if required
bool TFSKAnalyzeThread::CheckTerm()
{
    if (Terminated)
    {
        Term = true;

        // Stop indexing thread, if active
        if (IdxInfo.IsActive)
            m2IndexStop(&IdxInfo);

        // Reset app title
        Application->Title = FSKAppTitle;

        // Terminate
        return true;
    }
    else
        return false;
}

// Create the directory structure of the job
bool TFSKAnalyzeThread::ACreateJob()
{
    JobPath = FSKNewForm->JobPath;
    OutPath = FSKNewForm->OutPath;
    DVDPath = FSKNewForm->DVDPath;

    return GenerateDVDSum(DVDPath.c_str(), DVDSum) &&
    ForceDirectories(JobPath) && ForceDirectories(OutPath) && !FileSetAttr(JobPath, faHidden);
}


// Check if we have to import the index
bool TFSKAnalyzeThread::ACheckImport()
{
    int            i;             // counter
    FILE          *File;          // file handle
    char           buf[4096];     // character buffer
    unsigned long  _DVDSum;       // current DVDSum
    __int64        Size;          // current file's size
    AnsiString     Text;          // temporary string
    bool           OK;            // OK flag

    // Try all pending jobs
    for (i = 0; i < FSKMenuForm->JobList->Items->Count; i++)
        if (FSKMenuForm->JobList->Items->Item[i]->StateIndex == JS_Pending)
        {
            ImportPath = FSKMenuForm->JobList->Items->Item[i]->SubItems->Strings[JI_Path] + FSKJ_PATH;

            // Parse .idxlog file
            Text = ImportPath + FSKJ_IDXLOG_ID;
            if (!FileExists(Text) || (File = fopen(Text.c_str(), "rt")) == NULL)
                continue;

            fscanf(File, "%[^\n]\n", buf);
            if (strcmp(buf, FSKJ_IDXLOG_ID))
            {
                fclose(File);
                continue;
            }

            fscanf(File, "%[^\n]\n", buf);
            if (strcmp(AnsiString(buf).UpperCase().c_str(), DVDPath.UpperCase().c_str()))
            {
                fclose(File);
                continue;
            }

            if ((fscanf(File, "%x", &_DVDSum) != 1) || (_DVDSum != DVDSum) || feof(File))
            {
                fclose(File);
                continue;
            }

            fscanf(File, "%[^\n]\n", buf);

            TotalSize = 0;
            OK = true;
            IdxLog->Clear();
            while (!feof(File))
            {
                fscanf(File, "%[^\n]\n", buf);
                IdxLog->Add(AnsiString(buf));

                if (*buf != '>')
                {
                    Text = ImportPath + buf;
                    if (FileExists(Text) && ((Size = FileSize(Text)) >= 0))
                        TotalSize += Size;
                    else
                    {
                        OK = false;
                        break;
                    }
                }
            }

            fclose(File);
            if (OK && TotalSize > 0)
                return Application->MessageBoxA("Do you want to import the analysis data from a pending job?", FSK_PRODUCT_NAME, MB_YESNO | MB_ICONASTERISK) == mrYes;
        }

    return false;
}

// Get total size of VOB files to be analyzed. It will we used to calculate the
// progress and the total remaining time.
bool TFSKAnalyzeThread::ATotalSize()
{
    TSearchRec sr;
    int Result;
    int ID, Number;

    TotalSize = 0;

    // Find every VTS_??_*.VOB (except VTS_??_0.VOB)
    Result = FindFirst(DVDPath + "VTS_??_*.VOB", faAnyFile ^ faDirectory ^ faVolumeID, sr);

    while (!Result)
    {
        if ((sscanf(sr.Name.UpperCase().c_str(), "VTS_%d_%d.VOB", &ID, &Number) == 2) && Number)
        {
            __int64 Size;

            if ((Size = FileSize(DVDPath + sr.Name)) == -1)
            {
                FindClose(sr);
                return false;
            }
            else
                TotalSize += Size;
        }

        // Next VOB
        Result = FindNext(sr);
    }

    // Return
    if (TotalSize > 0)
        return true;
    else
        return false;
}

// Process all streams
bool TFSKAnalyzeThread::AProcessStreams()
{
    TSearchRec sr;
    int Result, Count = 0;
    int ID;                        // ID of the current stream
    AnsiString IfoName, VobName, StreamName;

    DoneSize = 0;

    // Get size of all VOBs
    if (!ATotalSize())
    {
        FatalError("Invalid source", "Invalid DVD source!");
        return false;
    }

    // Get current time
    StartTime = Now();
    _LastTime = 0;

    // Find every VTS_??_0.IFO
    Result = FindFirst(DVDPath + "VTS_??_0.IFO", faAnyFile ^ faDirectory ^ faVolumeID, sr);

    while (!Result)
    {
        // Termination point
        if (CheckTerm())
        {
            FindClose(sr);
            return false;
        }

        // Parse stream ID
        IfoName = sr.Name.UpperCase();
        if (sscanf(IfoName.c_str(), "VTS_%d_0.IFO", &ID) == 1)
        {
            VobName.sprintf("VTS_%.2d_1.VOB", ID);
            StreamName.sprintf("VTS_%.2d", ID);

            // Verify
            if ((FileSize(DVDPath + IfoName) > 0) && (FileSize(DVDPath + VobName) > 0))
            {
                // Process it
                AddLogEvent(TFSKNewForm::LSI_Video, "Analyzing " + StreamName);

                if (AProcessStream(StreamName, ID))
                {
                    // Add to list
                    Count++;
                }
                else
                {
                    // Remove from VTS list
                    FSKNewForm->TitleInfo[ID].PGCNum = 0;

                    // Termination point
                    if (Term)
                    {
                        FindClose(sr);
                        return false;
                    }
                }

                // Adjust progress bar
                DoneSize += CurSize;
                SetProgress(int((DoneSize * 100) / TotalSize));
            }
            else
            {
                // Warning
                Warning(StreamName + " is invalid");
            }
        }

        // Next IFO
        Result = FindNext(sr);
    }

    FindClose(sr);

    if (!Count)
    {
        FatalError("No valid title sets found", "Unable to find a valid title set! The DVD is probably damaged or encrypted!");
        return false;
    }

    // Success
    return true;
}

// Import index of all streams
bool TFSKAnalyzeThread::AImportStreams()
{
    int        i;                // counter;
    int        ID, Count = 0;    // ID of the current stream
    AnsiString StreamName;
    __int64    CurSize;          // size of the current file

    DoneSize = 0;

    // Copy files
    for (i = 0; i < IdxLog->Count; i++)
    {
        // Termination point
        if (CheckTerm())
            return false;

        // Next VTS/file
        if (IdxLog->Strings[i][1] == '>')
        {
            if (Count && !AProcessStream(StreamName, ID))
                goto Error;

            sscanf(IdxLog->Strings[i].c_str(), "> VTS_%d", &ID);
            StreamName.sprintf("VTS_%.2d", ID);
            AddLogEvent(TFSKNewForm::LSI_Video, "Importing " + StreamName);
            Count++;
        }
        else
        {
            if (((CurSize = FileSize(ImportPath + IdxLog->Strings[i])) < 0) ||
                !CopyFile((ImportPath + IdxLog->Strings[i]).c_str(), (JobPath + IdxLog->Strings[i]).c_str(), FALSE))
                    goto Error;

            // Adjust progress bar
            DoneSize += CurSize;
            SetProgress(int((DoneSize * 100) / TotalSize));
        }
    }

    if (!AProcessStream(StreamName, ID))
        goto Error;

    if (!CopyFile((ImportPath + FSKJ_IDXLOG_ID).c_str(), (JobPath + FSKJ_IDXLOG_ID).c_str(), FALSE))
        goto Error;

    return true;

// Error
Error:
    FatalError("Import error", "Unable to import analysis data! You should reanalyze the DVD!");
    return false;
}

// Process a stream
bool TFSKAnalyzeThread::AProcessStream(AnsiString Name, int ID)
{
    AnsiString      IfoPath, VobPath;           // file paths
    AnsiString      VobList;                    // list of VOB files to be processed
    AnsiString      Prefix;                     // job file name prefix (ex: "vts01")
    AnsiString      Path;                       // a file path
    AnsiString      Text;                       // temporary string
    int             Number;                     // number of a VOB file
    int             StreamID;                   // ID of the current stream
    ifo_t          *ifo;                        // IFO "handle"
    const pgc_t    *ifo_pgci;                   // IFO PGC info
    ifo_audio_t    *ifo_audio;                  // IFO audio info
    ifo_subpic_t   *ifo_subpic;                 // IFO subpicture info
    ifo_cell_tab_t *ifo_cell_tab;               // IFO cell table
    int             i, pgc, x, num, pgcnum;     // counters
    int             tmp;                        // temp variable
    FILE            *File;                      // file handle
    TSearchRec      sr;                         // search record
    int             Result;                     // search result

    // Prepare paths
    IfoPath = DVDPath;
    IfoPath.cat_printf("VTS_%.2d_0.IFO", ID);
    Prefix.printf("vts%.2d", ID);

    if (!Import)
    {
        // Create list of VOBs
        int Count = 0;

        VobList = "";
        Number = 1;
        CurSize = 0;

        for (; ;)
        {
            // Create file name string
            VobPath = DVDPath;
            VobPath.cat_printf("VTS_%.2d_%d.VOB", ID, Number);

            if (FileExists(VobPath))
            {
                __int64 Size;

                if ((Size = FileSize(VobPath)) == -1)
                    goto Invalid;
                CurSize += Size;

                VobList = VobList + VobPath + "|";
                Number++;
                Count++;
            }
            else
                break;
        }

        // Check
        IdxInfo.IsCSS = 0;
        if (!Count || (CurSize <= 0))
            goto Invalid;

        // Cut last delimiter
        VobList.SetLength(VobList.Length() - 1);
    }

    // Open IFO file
    if (!(ifo = ifoOpen(IfoPath.c_str())))
        goto Invalid;

    if (Import)
    {
        char buf[256];

        // Read .idxinfo file
        if ((File = fopen((JobPath + Prefix + ".idxinfo").c_str(), "rb")) == NULL)
            goto Invalid;

        if ((fgets(buf, strlen(FSKJ_IDXINFO_ID) + 1, File) == NULL) ||
            strcmp(buf, FSKJ_IDXINFO_ID))
                goto Invalid;

        if (fread(&IdxInfo, sizeof(IdxInfo), 1, File) < 1)
            goto Invalid;

        fclose(File);
    }
    else
    {
        // Index substreams
        TDateTime CurTime;
        __int64 Elapsed;

        if (m2Index(VobList.c_str(), (JobPath + Prefix).c_str(), 1, 1, &IdxInfo))
        {
            ifoClose(ifo);
            goto Invalid;
        }

        // Loop
        do
        {
            // Termination point
            if (CheckTerm())
            {
                ifoClose(ifo);
                return false;
            }

            // Update progress bar
            SetProgress(int((DoneSize * 100 + (int)IdxInfo.Progress * CurSize) / TotalSize));

            // Update estimated remaining time
            if (_Progress)
            {
                CurTime = Now();
                Elapsed = SecondsBetween(CurTime, StartTime);
                SetTime((Elapsed * (100 - _Progress)) / _Progress);
            }

            // Wait 1 second
            Sleep(1000);
        } while (IdxInfo.IsActive);

        // Check result
        if (!IdxInfo.IsSuccess || IdxInfo.IsCSS)
        {
            ifoClose(ifo);
            goto Invalid;
        }
    }

    //
    // Get stream info
    //

    // Video
    switch (IdxInfo.Video.System)
    {
    case VS_PAL:
        FSKNewForm->TitleInfo[ID].Video.IsPAL = true;
        break;

    default:
        FSKNewForm->TitleInfo[ID].Video.IsPAL = false;
        break;
    }

    FSKNewForm->TitleInfo[ID].Video.Width  = (short)IdxInfo.Video.Width;
    FSKNewForm->TitleInfo[ID].Video.Height = (short)IdxInfo.Video.Height;

    switch (IdxInfo.Video.AspectRatio)
    {
    case VAR_4_3:
        FSKNewForm->TitleInfo[ID].Video.IsWide = false;
        break;

    case VAR_16_9:
        FSKNewForm->TitleInfo[ID].Video.IsWide = true;
        break;

    default:
        ifoClose(ifo);
        goto Invalid;
    }

    if (IdxInfo.Video.FrameRate < 1.0)
    {
        ifoClose(ifo);
        goto Invalid;
    }

    FSKNewForm->TitleInfo[ID].Video.Framerate       = IdxInfo.Video.FrameRate;
    FSKNewForm->TitleInfo[ID].Video.IsProgressive   = IdxInfo.Video.IsProgressive;
    FSKNewForm->TitleInfo[ID].Video.FrameTypePurity = IdxInfo.Video.FrameTypePurity;
    FSKNewForm->TitleInfo[ID].Video.IsTFF           = IdxInfo.Video.IsTFF;
    FSKNewForm->TitleInfo[ID].Video.FilmPurity      = IdxInfo.Video.FilmPurity;

    // Get audio info
    if ((ifo_audio = ifoGetAudio(ifo)) == NULL)
    {
        ifoClose(ifo);
        goto Invalid;
    }

    // Get subpicture info
    if ((ifo_subpic = ifoGetSPU(ifo)) == NULL)
    {
        free(ifo_audio->sub);
        free(ifo_audio);
        ifoClose(ifo);
        goto Invalid;
    }

    // PGC
    if (!(pgcnum = ifoGetNumPGCI(ifo)))
    {
        // Error
        free(ifo_audio->sub);
        free(ifo_audio);
        free(ifo_subpic->sub);
        free(ifo_subpic);
        ifoClose(ifo);
        goto Invalid;
    }
    FSKNewForm->TitleInfo[ID].PGCNum  = pgcnum;
    FSKNewForm->TitleInfo[ID].PGC     = new TFSKNewForm::TPGCInfo[pgcnum + 1];

    for (pgc = 1; pgc <= pgcnum; pgc++)
    {
        // Get cell table
        if (!(ifo_cell_tab = ifoGetCells(ifo, pgc - 1)))
        {
            // Error
            free(ifo_audio->sub);
            free(ifo_audio);
            free(ifo_subpic->sub);
            free(ifo_subpic);
            ifoClose(ifo);
            goto Invalid;
        }

        // Get info
        FSKNewForm->TitleInfo[ID].PGC[pgc].Length.hour    = ifo_cell_tab->time[0];
        FSKNewForm->TitleInfo[ID].PGC[pgc].Length.minute  = ifo_cell_tab->time[1];
        FSKNewForm->TitleInfo[ID].PGC[pgc].Length.second  = ifo_cell_tab->time[2];
        FSKNewForm->TitleInfo[ID].PGC[pgc].Length.frame_u = ifo_cell_tab->time[3];

        FSKNewForm->TitleInfo[ID].PGC[pgc].AngleNum       = ifo_cell_tab->anglenum;

        // No numframes & crop info
        FSKNewForm->TitleInfo[ID].PGC[pgc].Video.NumFrames = -1;
        FSKNewForm->TitleInfo[ID].PGC[pgc].Video.Crop      = -1;

        // Free cell table
        free(ifo_cell_tab->cells);
        free(ifo_cell_tab);

        // Get PGC info
        if ((ifo_pgci = ifoGetPGCI(ifo, pgc - 1)) == NULL)
        {
            free(ifo_audio->sub);
            free(ifo_audio);
            free(ifo_subpic->sub);
            free(ifo_subpic);
            ifoClose(ifo);
            goto Invalid;
        }

        // Audio
        FSKNewForm->TitleInfo[ID].PGC[pgc].Audio = new TFSKNewForm::TAudioInfo[ifo_audio->num];

        num = 0;
        for (i = 0; i < (int)ifo_audio->num; i++)
        {
            // Get stream ID
            if (ifo_pgci->audio_status[i].available)
                StreamID = ifo_pgci->audio_status[i].link;
            else
                continue;

            // Check stream
            if (IdxInfo.Audio.Track[StreamID].Format == AF_None ||
                (IdxInfo.Audio.Track[StreamID].Format == AF_LPCM &&
                 IdxInfo.Audio.Track[StreamID].Quant != 16 && IdxInfo.Audio.Track[StreamID].Quant != 20 &&
                 IdxInfo.Audio.Track[StreamID].Quant != 24 && IdxInfo.Audio.Track[StreamID].Quant != 32))
                continue;

            // Index
            FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Index = StreamID;

            // Format
            FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Format = IdxInfo.Audio.Track[StreamID].Format;

            // Parameters
            if (IdxInfo.Audio.Track[StreamID].Format == AF_LPCM)
            {
                FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Channels   = IdxInfo.Audio.Track[StreamID].Channels;
                FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Quant      = IdxInfo.Audio.Track[StreamID].Quant;
                FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Samplerate = IdxInfo.Audio.Track[StreamID].Samplerate;
            }
            else
                FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Channels   = (char)ifo_audio->sub[i].num_channels;

            FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Delay    = (short)IdxInfo.Audio.Track[StreamID].Delay;
            FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Language = ifo_audio->sub[i].lang_code;
            FSKNewForm->TitleInfo[ID].PGC[pgc].Audio[num].Caption  = (char)ifo_audio->sub[i].caption;

            // Next
            num++;
        }

        FSKNewForm->TitleInfo[ID].PGC[pgc].AudioNum = num;

        // Subpicture
        FSKNewForm->TitleInfo[ID].PGC[pgc].Subpic = new TFSKNewForm::TSubpicInfo[ifo_subpic->num];

        num = 0;
        for (i = 0; i < (int)ifo_subpic->num; i++)
        {
            // Get stream ID
            if (ifo_pgci->subp_status[i].available)
            {
                if (FSKNewForm->TitleInfo[ID].Video.IsWide)
                    StreamID = ifo_pgci->subp_status[i].wide;
                else
                    StreamID = ifo_pgci->subp_status[i].format4_3;
            }
            else
                continue;

            // Check stream
            if (!IdxInfo.Subpic.Track[StreamID])
                continue;

            // Index
            FSKNewForm->TitleInfo[ID].PGC[pgc].Subpic[num].Index = StreamID;

            // Parameters
            FSKNewForm->TitleInfo[ID].PGC[pgc].Subpic[num].Language = ifo_subpic->sub[i].lang_code;

            // Next
            num++;
        }

        FSKNewForm->TitleInfo[ID].PGC[pgc].SubpicNum = num;
    }

    // Subpicture additional info
    for (i = 0; i < IFO_NUM_SPU; i++)
    {
        if (IdxInfo.Subpic.Track[i])
        {
            Text = JobPath + Prefix;
            Text.cat_printf("_st%d.sub", i);

            FSKNewForm->TitleInfo[ID].SubFilename[i] = Prefix;
            FSKNewForm->TitleInfo[ID].SubFilename[i].cat_printf("_pgc00_a0_st%d.sub", i);

            if (FileExists(Text) && !CopyFile(Text.c_str(), (JobPath + FSKNewForm->TitleInfo[ID].SubFilename[i]).c_str(), FALSE))
            {
                free(ifo_audio->sub);
                free(ifo_audio);
                free(ifo_subpic->sub);
                free(ifo_subpic);
                ifoClose(ifo);
                goto Invalid;
            }
        }
    }

    // Free memory
    free(ifo_audio->sub);
    free(ifo_audio);
    free(ifo_subpic->sub);
    free(ifo_subpic);

    // Close IFO file
    ifoClose(ifo);

    if (!Import)
    {
        // Write .idxinfo file
        if ((File = fopen((JobPath + Prefix + ".idxinfo").c_str(), "wb")) != NULL)
        {
            fputs(FSKJ_IDXINFO_ID, File);
            fwrite((const void *)&IdxInfo, sizeof(IdxInfo), 1, File);
            fclose(File);
        }

        // Append .idxlog file
        Text.printf("> VTS_%.2d\n", ID);
        IdxLog->Add(Text);

        // Find vts??*.* files
        Result = FindFirst(JobPath + Prefix + "*.*", faAnyFile ^ faDirectory ^ faVolumeID, sr);

        while (!Result)
        {
            if (!strstr(sr.Name.c_str(), "pgc00_a0"))
                IdxLog->Add(sr.Name + "\n");

            // Next file
            Result = FindNext(sr);
        }

        FindClose(sr);
    }

    // Success
    return true;

    // Error
Invalid:
    if (!Import)
    {
        if (IdxInfo.IsCSS)
            Warning(Name + " is CSS encrypted");
        else
            Warning(Name + " is invalid");
    }
    return false;
}

/*-----------------------------------------------------------------------------
    TFSKAnalyzeThread >> Form control methods
-----------------------------------------------------------------------------*/

// Add a log event
void TFSKAnalyzeThread::AddLogEvent(ELogStateIcon StateIcon, AnsiString Event)
{
    _LogStateIcon = StateIcon;
    _LogEvent = Event;

    Synchronize(_AddLogEvent);
}

// Show info
void TFSKAnalyzeThread::Info(AnsiString Text)
{
    AddLogEvent(TFSKNewForm::LSI_Info, Text);
}

// Show warning
void TFSKAnalyzeThread::Warning(AnsiString Text)
{
    AddLogEvent(TFSKNewForm::LSI_Warning, Text);
}

void __inline Warning(AnsiString Text);

// Show an error message
void TFSKAnalyzeThread::Error(AnsiString Text1, AnsiString Text2)
{
    _LogEvent = Text1;

    Synchronize(_Error);

    // Show message box
    Application->MessageBoxA(Text2.c_str(), FSK_PRODUCT_NAME, MB_ICONERROR | MB_OK);
}

// Show an error message and terminate thread
void TFSKAnalyzeThread::FatalError(AnsiString Text1, AnsiString Text2)
{
    // Delete job directory, if created
    if (DirectoryExists(FSKNewForm->JobRoot))
        RemoveDirTree(FSKNewForm->JobRoot);

    // Error
    Error(Text1, Text2);
    Finish(false);
}

// Send finish message to the form
void TFSKAnalyzeThread::Finish(bool Success)
{
    _Success = Success;

    Synchronize(_Finish);
}

// Set progress bar position
void __inline TFSKAnalyzeThread::SetProgress(int Progress)
{
    _Progress = Progress;

    Synchronize(_SetProgress);
}

// Set remaining time text
void __inline TFSKAnalyzeThread::SetTime(int Seconds)
{
    _Time = Seconds;

    Synchronize(_SetTime);
}

/*-----------------------------------------------------------------------------
    TFSKAnalyzeThread >> Thread-safe methods
-----------------------------------------------------------------------------*/

// Add a log event
void __fastcall TFSKAnalyzeThread::_AddLogEvent()
{
    FSKNewForm->AddLogEvent(_LogStateIcon, _LogEvent);
}

// Show an error message
void __fastcall TFSKAnalyzeThread::_Error()
{
    FSKNewForm->AddLogEvent(TFSKNewForm::LSI_Error, _LogEvent);
}

// Send finish message to the form
void __fastcall TFSKAnalyzeThread::_Finish()
{
    // Reset app title
    Application->Title = FSKAppTitle;

    FSKNewForm->S2_Finish(_Success);
}

// Set progress bar position
void __fastcall TFSKAnalyzeThread::_SetProgress()
{
    AnsiString Title;

    // Adjust progress bar
    FSKNewForm->ProgressBar->Position = _Progress;

    // Update application title
    Title = FSKAppTitle;
    Title.cat_printf(" - %d%%", _Progress);
    Application->Title = Title;
}

// Set remaining time text
void __fastcall TFSKAnalyzeThread::_SetTime()
{
    AnsiString Text;

    // Hide labels if finished
    if (!_Time)
    {
        FSKNewForm->TimeTextLabel->Visible = false;
        FSKNewForm->TimeLabel->Visible = false;
        return;
    }

    // Return if the prediction is not yet accurate
    if (!FSKNewForm->TimeLabel->Visible && abs(_LastTime - _Time) > 30)
    {
        _LastTime = _Time;
        return;
    }

    // Show time
    if (_Time < 60)
    {
        if (_Time == 1)
            Text = Text.sprintf("1 second", _Time);
        else
            Text = Text.sprintf("%d seconds", _Time);
    }
    else
    {
        _Time = _Time / 60;

        if (_Time == 1)
            Text = Text.sprintf("1 minute", _Time);
        else
            Text = Text.sprintf("%d minutes", _Time);

        _LastTime = _Time;
    }

    _LastTime = _Time;

    // Set caption
    FSKNewForm->TimeLabel->Caption = Text;

    // Show time label if not visible
    if (!FSKNewForm->TimeTextLabel->Visible)
    {
        FSKNewForm->TimeTextLabel->Visible = true;
        FSKNewForm->TimeLabel->Visible = true;
    }
}

/*=============================================================================
    TFSKSelectThread
=============================================================================*/

/*-----------------------------------------------------------------------------
    TFSKSelectThread >> Constructors
-----------------------------------------------------------------------------*/

// Default constructor
__fastcall TFSKSelectThread::TFSKSelectThread() : TThread(true)
{
    // Selfconfigure
    FreeOnTerminate = true;
}

/*-----------------------------------------------------------------------------
    TFSKSelectThread >> Main methods
-----------------------------------------------------------------------------*/

void __fastcall TFSKSelectThread::Execute()
{
    int        VTS, PGC, Angle; // selected title info
    AnsiString ReIndexPath;     // rebuilt index path
    AnsiString IndexName;       // name/path of the original index
    AnsiString IfoPath;         // path of the IFO file
    bool       RebuildVideo;    // do we have to rebuild video?
    int        _Width, _Height;

    // Get title info
    VTS   = FSKNewForm->VTS;
    PGC   = FSKNewForm->PGC;
    Angle = FSKNewForm->Angle;

    FSKNewForm->ReIndexName.printf("vts%.2d_pgc%.2d_a%d", VTS, PGC, Angle);
    ReIndexPath = FSKNewForm->JobPath + FSKNewForm->ReIndexName;

    // Check if we have to rebuild video index
    RebuildVideo = !FileExists(ReIndexPath + ".m2vi");

    // Rebuild index with audio
    IndexName = FSKNewForm->JobPath;
    IndexName.cat_printf("vts%.2d", VTS);
    IfoPath = FSKNewForm->DVDPath;
    IfoPath.cat_printf("VTS_%.2d_0.IFO", VTS);

    if (!m2Rebuild(IndexName.c_str(), IfoPath.c_str(), ReIndexPath.c_str(), RebuildVideo, true, true, PGC, Angle))
    {
        // Show error message
        Application->MessageBoxA("Unable to process selected title! The DVD is probably damaged!", FSK_PRODUCT_NAME, MB_OK | MB_ICONERROR);

        // Close forms
        Synchronize(_Error);
        return;
    }

    // Get number of frames
    IndexName = FSKNewForm->JobPath + FSKNewForm->ReIndexName;
    IndexName.cat_printf(".m2vi");

    if (!m2PreviewOpen(IndexName.c_str(), _Width, _Height, FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.NumFrames, false))
    {
        // Show error message
        Application->MessageBoxA("Frame count detection failed! The DVD is probably damaged!", FSK_PRODUCT_NAME, MB_OK | MB_ICONERROR);

        // Close forms
        Synchronize(_Error);
        return;
    }

    m2PreviewClose();

    // Crop borders
    if (!m2Crop(IndexName.c_str(), FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Left,
                                   FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Top,
                                   FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Width,
                                   FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Height,
                                   int(FSKNewForm->TitleInfo[VTS].Video.FilmPurity) > 0))
    {
        // Show error message
        Application->MessageBoxA("Border detection failed! The DVD is probably damaged!", FSK_PRODUCT_NAME, MB_OK | MB_ICONERROR);

        // Close forms
        Synchronize(_Error);
        return;
    }

    // Adjustment
    FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Top    = (FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Top / 2) * 2;
    FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Left   = (FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Left / 2) * 2;
    FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Width  = (FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Width / 2) * 2;
    FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Height = (FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Height / 2) * 2;

    if (!FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Top &&
        !FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Left &&
        FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Width == FSKNewForm->TitleInfo[VTS].Video.Width &&
        FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Height == FSKNewForm->TitleInfo[VTS].Video.Height)
    {
        // No crop
        FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Crop = 0;
    }
    else
    {
        // Crop
        FSKNewForm->TitleInfo[VTS].PGC[PGC].Video.Crop = 1;
    }

    // Finish
    Synchronize(_Finish);
}

/*-----------------------------------------------------------------------------
    TFSKSelectThread >> Thread-safe methods
-----------------------------------------------------------------------------*/

// Success
void __fastcall TFSKSelectThread::_Finish()
{
    FSKNewWaitForm->Close();
    FSKNewForm->NextButtonClick(this);
}

void __fastcall TFSKSelectThread::_Error()
{
    FSKNewForm->TitleInfo[FSKNewForm->VTS].PGC[FSKNewForm->PGC].Video.NumFrames = -1;
    FSKNewForm->S3_PrevVTS = FSKNewForm->S3_PrevPGC = FSKNewForm->S3_PrevAngle = -1; 

    // Close wait form
    FSKNewWaitForm->Close();
    FSKNewForm->S3_Begin();
}
