/*=============================================================================
    Project Name:
            FilmShrink

	Module Name:
			FSKTemplate.h

	Abstract:
			FilmShrink templates and configuration

    License:
            GNU General Public License (GPL)

	Revision History:
			* Oct 27, 2004:  Created by Attila Afra
            * Nov 13, 2004:  RealProducer Job File support
            * Jan  7, 2005:  Subpicture support
            * Feb  6, 2005:  Bitrate overhead added
            * Feb 17, 2005:  Loss protection support
            * Feb 22, 2005:  Advanced video codec params

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

#ifndef FSKTemplateH
#define FSKTemplateH

#pragma warn -8006  // Initializing enum with int
#pragma warn -8058  // Pre-compiled header

#include "m2index_info.h"

/*-----------------------------------------------------------------------------
    General
-----------------------------------------------------------------------------*/

// Volume size presets
const struct
{
    char Name[32];    // name of the preset
    int  Size;        // volume size
    int  Type;        // volume type (1 - Disk)
} FSKVolumeSize[] = {
                                              // Custom
                        "CD [74 min]", 650, 1,  // 650 MB CD
                        "CD [80 min]", 700, 1,  // 700 MB CD

                        0,           0,   0
                    };

#define FSK_DEFAULT_VOLUME_IDX    2  // 700 MB CD

// Output
#define FSK_OUT_FILENAME            "out%.2d"             // name of a job file w/ volume index (formatted)
#define FSKJ_ID                     ".fskj"               // FilmShrink job ID
#define FSKJ_PATH                   "\\"FSKJ_ID"\\"       // job directory path
#define FSKJ_INFO_FILENAME          FSKJ_PATH"\\"FSKJ_ID  // name of the job info file
#define FSKJ_AVS_DEBUG_FILENAME     "avsdebug"
#define FSKJ_IDXINFO_ID             "m2IndexInfo"         // .idxinfo file ID
#define FSKJ_IDXLOG_ID              ".idxlog"             // .idxlog file ID

#define FSK_OVERHEAD                0.9905                // bitrate/size overhead coeff

/*-----------------------------------------------------------------------------
    Video
-----------------------------------------------------------------------------*/

// Video compression params
#define FSK_VIDEO_BPP           0.225           // optimal video quality in BPP
#define FSK_MIN_VIDEO_BITRATE   8               // minimum video bitrate (Kbps)

#define FSK_MIN_VIDEO_RES       32              // minimum video width/height
#define FSK_MAX_VIDEO_RES       2048            // maximum video width/height

#define FSK_VIDEO_DECODER       "m2video.dll"   // MPEG-2 video decoder plugin
#define FSK_VIDEO_DECOMB        "Decomb.dll"

// Relative video compression speed (depending on EHQ) - the higher the slower!
#define FSK_EHQ_100     100  // Very High   (95-100)
#define FSK_EHQ_90      60   // High        (85-94)
#define FSK_EHQ_80      45   // Medium-High (78-84)
#define FSK_EHQ_70      33   // Medium      (70-77)
#define FSK_EHQ_60      26   // Low         (60-69)
#define FSK_EHQ_50      22   // Very Low    (50-59)

// Video complexity preset type
struct TFSKVideoComplexity
{
    int Encoding;   // encoding pass
    int Analysis;   // analysis pass (-1 = 1 pass)
};

// Video complexity preset in bitrate mode
const TFSKVideoComplexity FSKVideoComplexityBitrate[] = {
    // Very low
    50, -1,
    60, -1,

    // Low
    70, -1,
    80, -1,

    // Medium
    60, 50,
    70, 50,

    // High
    80, 50,
    90, 50,

    // Very high
    100, 50,

    -1, -1
};

// Video complexity preset in quality mode
const TFSKVideoComplexity FSKVideoComplexityQuality[] = {
    // Very low
    50, -1,

    // Low
    60, -1,

    // Medium
    70, -1,
    80, -1,

    // High
    90, -1,

    // Very high
    100, -1,

    -1, -1
};

/*-----------------------------------------------------------------------------
    Audio
-----------------------------------------------------------------------------*/

// Audio codecs
const struct
{
    char Name[32];          // name of the codec
    char DecoderName[256];  // filename of the decoder plugin
    char FuncName[256];     // function name
    int  IconIndex;         // icon index
} FSKAudioCodecs[] = {
                        "Unknown",       "",              "",          -1, // AF_None
                        "Dolby Digital", "m2audio_ac3.dll",  "m2AudioAC3Source",  0, // AF_AC3
                        "DTS",           "m2audio_dts.dll",  "m2AudioDTSSource",  1, // AF_DTS
                        "MPEG-1",        "m2audio_mpa.dll",  "m2AudioMPASource",  2, // AF_MPA
                        "LPCM",          "m2audio_lpcm.dll", "m2AudioLPCMSource", 3, // AF_LPCM
                     };

// Recommanded audio/total bitrate ratio
#define FSK_AUDIO_BITRATE_RATIO     0.15

// Audio channel mode type
enum EFSKAudioMode
{
    ch1  = 0,   // Mono
    ch2,        // Stereo
    ch2x,       // Surround
    chM         // Multichannel
};

// Audio channel mode names
const char FSKAudioModeNames[][32] = {"Mono", "Stereo", "Surround", "Multichannel"};

// Audio preset type
struct TFSKAudioPreset
{
    int           Bitrate;   // bandwidth (in Kbps)
    EFSKAudioMode Mode;      // channel mode

    char          Type[5];   // codec name (FOURCC)
    int           Flavor;    // codec flavor
};

// Mono audio presets
const TFSKAudioPreset FSKAudioMonoPresets[] =
{
     6, ch1, "cook",  8,  //  6 Kbps Music - RealAudio
     8, ch1, "cook",  0,  //  8 Kbps Music - RealAudio
    11, ch1, "cook",  1,  // 11 Kbps Music - RealAudio
    16, ch1, "cook",  2,  // 16 Kbps Music - RealAudio
    20, ch1, "cook", 15,  // 20 Kbps Music High Response - RealAudio
    32, ch1, "cook", 16,  // 32 Kbps Music High Response - RealAudio
    44, ch1, "cook",  5,  // 44 Kbps Music - RealAudio
    64, ch1, "cook",  6,  // 64 Kbps Music - RealAudio

     0,   0, 0,  0
};

// Stereo audio presets
const TFSKAudioPreset FSKAudioStereoPresets[] =
{
     12, ch2, "cook", 26,  //  12 Kbps Stereo Music - RealAudio
     16, ch2, "cook", 17,  //  16 Kbps Stereo Music - RealAudio
     20, ch2, "cook", 19,  //  20 Kbps Stereo Music High Response - RealAudio
     32, ch2, "racp",  4,  //  32 Kbps Stereo Music - RealAudio 10 (aacPlus)
     48, ch2, "racp",  3,  //  48 Kbps Stereo Music - RealAudio 10 (aacPlus)
     64, ch2, "racp",  0,  //  64 Kbps Stereo Music - RealAudio 10 (aacPlus)
     96, ch2, "racp",  1,  //  96 Kbps Stereo Music - RealAudio 10 (aacPlus)
    128, ch2, "racp",  2,  // 128 Kbps Stereo Music - RealAudio 10 (aacPlus)
    160, ch2, "raac",  3,  // 160 Kbps Stereo Music - RealAudio 10
    192, ch2, "raac",  4,  // 192 Kbps Stereo Music - RealAudio 10
    256, ch2, "raac",  5,  // 256 Kbps Stereo Music - RealAudio 10
    320, ch2, "raac",  6,  // 320 Kbps Stereo Music - RealAudio 10

      0,   0, 0,  0
};

// Stereo surround audio presets
const TFSKAudioPreset FSKAudioSurroundPresets[] =
{
     12, ch2,  "cook", 26,  //  12 Kbps Stereo Music - RealAudio
     16, ch2,  "cook", 17,  //  16 Kbps Stereo Music - RealAudio
     20, ch2,  "cook", 19,  //  20 Kbps Stereo Music High Response - RealAudio
     32, ch2,  "racp",  4,  //  32 Kbps Stereo Music - RealAudio 10 (aacPlus)

     44, ch2x, "cook", 29,  //  44 Kbps Stereo Surround Audio - RealAudio
     64, ch2x, "cook", 27,  //  64 Kbps Stereo Surround Audio - RealAudio
     96, ch2x, "cook", 28,  //  96 Kbps Stereo Surround Audio - RealAudio
    128, ch2x, "raac",  7,  // 128 Kbps Stereo Surround - RealAudio 10
    160, ch2x, "raac",  8,  // 160 Kbps Stereo Surround - RealAudio 10
    192, ch2x, "raac",  9,  // 192 Kbps Stereo Surround - RealAudio 10
    256, ch2x, "raac", 10,  // 256 Kbps Stereo Surround - RealAudio 10
    320, ch2x, "raac", 11,  // 320 Kbps Stereo Surround - RealAudio 10

      0, 0,    0,  0
};

// Multichannel audio presets
const TFSKAudioPreset FSKAudioMultiPresets[] =
{
     96, chM, "cook", 30,  //  96 Kbps 5.1 Multichannel - RealAudio 10
    132, chM, "cook", 31,  // 132 Kbps 5.1 Multichannel - RealAudio 10
    184, chM, "cook", 32,  // 184 Kbps 5.1 Multichannel - RealAudio 10
    268, chM, "cook", 33,  // 268 Kbps 5.1 Multichannel - RealAudio 10

      0, 0,   0,  0
};

/*-----------------------------------------------------------------------------
    Subpicture
-----------------------------------------------------------------------------*/

#define FSK_SUBPIC_DECODER  "VSFilter.dll"   // VobSub subpicture decoder

/*-----------------------------------------------------------------------------
    Functions
-----------------------------------------------------------------------------*/

// Create an RPJF file
/*
bool FSKWriteRPJF(const char *Filename,
                  bool TwoPass,
                  char *Title, char *Author, char *Copyright, char *Keywords,
                  char *Desc, int Rating,
                  char *Input, char *Output,
                  int   OutputRollSize,
                  int   AvgBitrate, int MaxBitrate,
                  int   VideoEncComplexity, int VideoPass1Complexity,
                  float VideoMaxFrameRate,
                  int   VideoQuality,
                  bool  LossProtection,

                  bool  Audio,
                  char* AudioCodecName = 0, int AudioCodecFlavor = 0);
*/

bool FSKWriteRPJF(const char *Filename,

                  // Clip info
                  char *Title, char *Author, char *Copyright, char *Keywords,
                  char *Desc, int Rating,
                  char *Input, char *Output,

                  // General
                  bool  enableTwoPass,
                  int   destinationRollSize,
                  int   avgBitrate, int maxBitrate,

                  // Video
                  int   encoderComplexity, int firstPassComplexity,
                  bool  enableRPR,
                  bool  rcEnableCurveCompression,
                  int   rcHighBitrateReduce, int rcLowBitrateBoost,
                  int   rcPFrameRefQuant, int rcBFrameRefQuant,
                  bool  AdaptiveB,
                  int   maxConsecutiveBFrames, int patternAdaptivity,
                  int   inloopCutOffQuant,
                  bool  inloopCutOffCompatible, bool inloopCutOffBUseRefQuant,
                  bool  chromaModeDecision, bool encodeAllFrames,
                  int   quality,
                  float FrameRate,
                  bool  enableLossProtection,

                  // Audio
                  bool  Audio,
                  char *codecName = 0, int codecFlavor = 0);


// Get the video complexity from an RPJF file
bool FSKGetRPJFVideoComplexity(const char *Filename, bool& TwoPass, int& VideoEncComplexity, int& VideoPass1Complexity);

// Get relative video compression speed
int FSKVideoEHQSpeed(int EHQ);

#endif // FSKTemplateH
