/*=============================================================================
	Module Name:
			Shutdown.cpp

	Abstract:
			Windows Shutdown API

    License:
            GNU General Public License (GPL)

	Revision History:
			* Aug 22, 2004:  Created by Attila Afra

------------------------------------------------------------------------------
	Copyright  2004 Attila Afra. All rights reserved.
=============================================================================*/

#include <windows.h>
#include "shutdown.h"

// Shuts down the system traditionally
bool Shutdown(EShutdownAction Action)
{
    HANDLE hToken;
    TOKEN_PRIVILEGES tkp;
    UINT Flags;

    // Get a token for this process
    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
        return false;

    // Get the LUID for the shutdown privilege
    LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);

    tkp.PrivilegeCount = 1;  // one privilege to set
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    // Get the shutdown privilege for this process
    AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES)NULL, 0);

    // Cannot test the return value of AdjustTokenPrivileges
    if (GetLastError() != ERROR_SUCCESS)
        return false;

    // Shut down the system
    switch (Action)
    {
    case SA_PowerOff:
        Flags = EWX_POWEROFF;
        break;

    case SA_Reboot:
        Flags = EWX_REBOOT;
        break;

    case SA_LogOff:
        Flags = EWX_LOGOFF;
        break;

    default:
        return false;
    }

    if (!ExitWindowsEx(Flags, 0))
        return false;

    return true;
}

/*----------------------------------------------------------------------------*/

typedef enum _SHUTDOWN_ACTION
{
    ShutdownNoReboot,
	ShutdownReboot,
	ShutdownPowerOff
} SHUTDOWN_ACTION;

typedef NTSYSAPI DWORD (NTAPI *NtShutdownSystemFunc)(IN SHUTDOWN_ACTION Action);


// Immediately shuts down the system
// WARNING: Forces processes to terminate, so this can cause the applications to
//          lose data. Therefore, you should only use this flag in an emergency.
bool FastShutdown(EShutdownAction Action)
{
	HINSTANCE hDLL = LoadLibrary("ntdll.dll");
	HANDLE hToken;
	TOKEN_PRIVILEGES tkp;
	NtShutdownSystemFunc pf;
    SHUTDOWN_ACTION Flag;
    UINT Param;

    // Check if the DLL is loaded
	if (!hDLL)
        return false;

    // Get a token for this process
    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
        return false;

    // Get the LUID for the shutdown privilege
    LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);

    tkp.PrivilegeCount = 1;  // one privilege to set
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    // Get the shutdown privilege for this process
    AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES)NULL, 0);

    // Cannot test the return value of AdjustTokenPrivileges
    if (GetLastError() != ERROR_SUCCESS)
        return false;

    // Get shutdown function address
	pf = (NtShutdownSystemFunc) GetProcAddress(hDLL, "NtShutdownSystem");
	if (!pf)
        return false;

    // Shut down the system (fast!)
    switch (Action)
    {
    case SA_PowerOff:
        Flag = ShutdownPowerOff;
        break;

    case SA_Reboot:
        Flag = ShutdownReboot;
        break;

    default:
        return false;
    }

    Param = (unsigned char)Flag;
    pf((SHUTDOWN_ACTION)Param);

	FreeLibrary(hDLL);
    return true;
}
